from copy import deepcopy

import mne
import numpy as np
from joblib import Parallel, delayed


def make_epochs(z_hat, info, t_lim, n_times_atom=1):
    """Make Epochs on the activations of atoms.
    n_splits, n_atoms, n_times_valid = z_hat.shape
    n_trials, n_atoms, n_times_epoch = z_hat_epoch.shape
    """
    n_splits, n_atoms, n_times_valid = z_hat.shape
    n_times = n_times_valid + n_times_atom - 1
    # pad with zeros
    padding = np.zeros((n_splits, n_atoms, n_times_atom - 1))
    z_hat = np.concatenate([z_hat, padding], axis=2)
    # reshape into an unique time-serie per atom
    z_hat = np.reshape(z_hat.swapaxes(0, 1), (n_atoms, n_splits * n_times))

    # create trials around the events, using mne
    new_info = mne.create_info(ch_names=n_atoms, sfreq=info['sfreq'])
    rawarray = mne.io.RawArray(data=z_hat, info=new_info, verbose=False)
    t_min, t_max = t_lim
    epochs = mne.Epochs(rawarray, info['events'], info['event_id'], t_min,
                        t_max, verbose=False)
    z_hat_epoched = epochs.get_data()
    return z_hat_epoched


def make_evoke(array, info, t_lim):
    """Compute evoked activations"""
    if array.ndim == 1:
        array = array[None, None]
    elif array.ndim == 2:
        array = array[None]
    epoched_array = make_epochs(array, info, t_lim=t_lim)
    evoked_array = epoched_array.mean(axis=0)
    return evoked_array


def make_evoke_one_surrogate(array, info, t_lim):
    # generate random events
    info = deepcopy(info)
    n_events = info['events'].shape[0]
    events = np.random.randint(array.shape[-1], size=n_events)
    events = np.sort(np.unique(events))
    n_events = events.shape[0]

    event_id = np.atleast_1d(info['event_id']).astype('int')
    n_tile = int(np.ceil(n_events / float(event_id.shape[0])))
    event_id_tiled = np.tile(event_id, n_tile)[:n_events]

    events = np.c_[events, np.zeros_like(events), event_id_tiled]
    info['events'] = events

    # make evoked with random events
    evoked_array = make_evoke(array, info, t_lim)
    return evoked_array


def make_evoke_all_surrogates(array, info, t_lim, n_jobs, n_surrogates=100):
    delayed_func = delayed(make_evoke_one_surrogate)
    evoked_arrays = Parallel(n_jobs=n_jobs)(delayed_func(array, info, t_lim)
                                            for i in range(n_surrogates))
    return np.array(evoked_arrays)


def plot_evoked_surrogates(array, info, t_lim, ax, n_jobs, label='',
                           threshold=0.005):
    """Compute and plot evoked array distribution over random events"""
    assert array.ndim == 1
    assert ax is not None
    # compute mean over epochs
    evoked = make_evoke(array, info, t_lim)[0]

    # compute surrogate evoked with random events
    evoked_surrogate = make_evoke_all_surrogates(array, info, t_lim,
                                                 n_jobs)[:, 0]

    # find thresholds
    low, high = 100 * threshold / 2., 100 * (1 - threshold / 2.)
    threshold_low = np.percentile(evoked_surrogate.min(axis=1), low)
    threshold_high = np.percentile(evoked_surrogate.max(axis=1), high)

    # plot the evoked and a gray area for the 95% percentile
    t = np.arange(len(evoked)) / info['sfreq'] + t_lim[0]
    outside_thresholds = ((evoked > threshold_high) + (evoked < threshold_low))
    color = 'C1' if np.any(outside_thresholds) else 'C2'
    ax.plot(t, evoked, label=label, color=color)
    label_th = str(100 * (1 - threshold)) + ' %'
    ax.fill_between(t, threshold_low, threshold_high, color='k', alpha=0.2,
                    label=label_th)
    ax.fill_between(t, threshold_low, threshold_high, where=outside_thresholds,
                    color='y', alpha=0.2)
    ax.axvline(0, color='k', linestyle='--')
    ax.set_ylim([0, None])
    ax.legend()

    # # plot the histogram of evoked_surrogate, and of evoked
    # ax = axes[1]
    # ax.hist(evoked_surrogate, bins=100, density=True, label='surrogate')
    # ax.hist(evoked.ravel(), bins=100, density=True, alpha=0.8,
    #         label='evoked')
    # ax.axvline(threshold_low, color='k', linestyle='--')
    # ax.axvline(threshold_high, color='k', linestyle='--')
    # ax.legend()



# MIT License
#
# Copyright (c) 2016 Olivier Bachem
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Fast and Provably Good Seedings for k-Means using k-MC2 and AFK-MC2

Cython implementation of the algorithms of the following two papers:

> Approximate K-Means++ in Sublinear Time
> Olivier Bachem, Mario Lucic, S. Hamed Hassani and Andreas Krause
> In Proc. Conference on Artificial Intelligence (AAAI), 2016.

> Fast and Provably Good Seedings for k-Means
> Olivier Bachem, Mario Lucic, S. Hamed Hassani and Andreas Krause
> To appear in Neural Information Processing Systems (NIPS), 2016.

Usage:
>>> import kmc2
>>> X = <Numpy array containing the data>
>>> seeding = kmc2.kmc2(X, 5)  # Run k-MC2 with k=5

Afterwards, we may run MiniBatchKMeans of Scikit-Learn:
>>> from sklearn.cluster import MiniBatchKMeans
>>> model = MiniBatchKMeans(5, init=seeding).fit(X)
>>> new_centers = model.cluster_centers_

Please refer to the doc string of kmc2.kmc2 for detailed usage.
"""
import numpy as np
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.utils import check_array
from sklearn.utils.validation import check_random_state
cimport numpy as np
cimport cython

from custom_distances import roll_invariant_euclidean_distances
from custom_distances import translation_invariant_euclidean_distances


def kmc2(X, k, chain_length=200, afkmc2=True, random_state=None, weights=None,
         distances='euclidean'):
    """Cython implementation of k-MC2 and AFK-MC2 seeding

    Args:
      X: (n,d)-shaped np.ndarray with data points (or scipy CSR matrix)
      k: number of cluster centers
      chain_length: length of the MCMC chain
      afkmc2: Whether to run AFK-MC2 (if True) or vanilla K-MC2 (if False)
      random_state: numpy.random.RandomState instance or integer to be used as seed
      weights: n-sized np.ndarray with weights of data points (default: uniform weights)
      distances: in {'euclidean', 'roll_inv', 'trans_inv'}

    Returns:
      centers: (k, d)-shaped numpy.ndarray with cluster centers
      centers_indices: (k, )-shaped numpy.ndarray with cluster centers indices
    """
    # Local cython variables
    cdef np.intp_t j, curr_ind
    cdef double cand_prob, curr_prob
    cdef double[::1] q_cand, p_cand, rand_a

    # Handle input
    X = check_array(X, accept_sparse="csr", dtype=np.float64, order="C")
    sparse = not isinstance(X, np.ndarray)
    if weights is None:
        weights = np.ones(X.shape[0], dtype=np.float64)
    random_state = check_random_state(random_state)

    if distances == 'euclidean':
        distances = euclidean_distances
    elif distances == 'roll_inv':
        distances = roll_invariant_euclidean_distances
    elif distances == 'trans_inv':
        distances = translation_invariant_euclidean_distances
    else:
        raise ValueError('Unknown distances: "%s".' % (distances, ))

    # Initialize result
    centers_indices = np.zeros((k, ), np.intp)
    centers = np.zeros((k, X.shape[1]), np.float64, order="C")
    # Sample first center and compute proposal
    centers_indices[0] = random_state.choice(X.shape[0], p=weights/weights.sum())
    rel_row = X[centers_indices[0], :]
    centers[0, :] = rel_row.todense().flatten() if sparse else rel_row
    if afkmc2:
        di = np.min(distances(X, centers[0:1, :], squared=True), axis=1)*weights
        q = di/np.sum(di) + weights/np.sum(weights)  # Only the potentials
    else:
        q = np.copy(weights)
    # Renormalize the proposal distribution
    q = q / np.sum(q)

    for i in range(k-1):
        # Draw the candidate indices
        cand_ind = random_state.choice(X.shape[0], size=(chain_length), p=q).astype(np.intp)
        # Extract the proposal probabilities
        q_cand = q[cand_ind]
        # Compute pairwise distances
        dist = distances(X[cand_ind, :], centers[0:(i+1), :], squared=True)
        # Compute potentials
        p_cand = np.min(dist, axis=1)*weights[cand_ind]
        # Compute acceptance probabilities
        rand_a = random_state.random_sample(size=(chain_length))
        with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
            # Markov chain
            for j in range(q_cand.shape[0]):
                cand_prob = p_cand[j]/q_cand[j]
                if j == 0 or curr_prob == 0.0 or cand_prob/curr_prob > rand_a[j]:
                    # Init new chain             Metropolis-Hastings step
                    curr_ind = j
                    curr_prob = cand_prob
        centers_indices[i+1] = cand_ind[curr_ind]
        rel_row = X[centers_indices[i+1], :]
        centers[i+1, :] = rel_row.todense().flatten() if sparse else rel_row
    return centers, centers_indices



import numpy as np
from scipy.signal import correlate
from scipy.linalg import eigh, blas

from alphacsc.utils import check_random_state


def learn_atoms(X, n_atoms, n_times_atom, n_iter=10, max_shift=11,
                random_state=None):
    """Learn atoms using the MoTIF algorithm.

    Parameters
    ----------
    X : array, shape (n_trials, n_times)
        The data on which to apply MoTIF.
    n_atoms : int
        The number of atoms.
    n_times_atom : int
        The support of the atoms
    n_iter : int
        The number of iterations
    max_shift : int
        The maximum allowable shift for the atoms.
    random_state : int | None
        The random initialization.
    """

    rng = check_random_state(random_state)

    n_trials, n_times = X.shape

    atoms = rng.rand(n_atoms, n_times_atom)
    corrs = np.zeros(n_trials)
    match = np.zeros((n_atoms, n_trials), dtype=np.int)

    # loop through atoms
    for k in range(n_atoms):

        aligned_data = np.zeros((n_times_atom, n_trials))

        # compute Bk
        B = np.zeros((n_times_atom, n_times_atom), order='F')
        for l in range(k):
            for p in np.arange(max_shift):
                atom_shifted = np.roll(atoms[l], -p)[np.newaxis, :]
                # B += np.dot(atom_shifted.T, atom_shifted)
                B = blas.dger(1, atom_shifted, atom_shifted, a=B,
                              overwrite_a=1)

        # make B invertible by adding a full-rank matrix
        B += np.eye(B.shape[0]) * np.finfo(np.float32).eps

        for i in range(n_iter):
            print('[seed %s] Atom %d Iteration %d' % (random_state, k, i))
            # loop through training data
            for n in range(n_trials):
                # ### STEP 1: Find out where the data and atom align ####

                # which of these to use for template matching?
                vec1 = (X[n] - np.mean(X[n])) / (np.std(X[n]) * len(X[n]))
                vec2 = (atoms[k] - np.mean(atoms[k])) / np.std(atoms[k])
                tmp = np.abs(correlate(vec1, vec2, 'same'))

                offset = n_times_atom // 2
                match[k, n] = tmp[offset:-offset].argmax() + offset
                corrs[n] = tmp[match[k, n]]

                # aligned_data[:, n] = np.roll(X[n], -shift[n])[:n_times_atom]
                aligned_data[:, n] = X[n, match[k, n] - offset:
                                       match[k, n] + offset].copy()

            # ### STEP 2: Solve the generalized eigenvalue problem ####
            A = np.dot(aligned_data, aligned_data.T).copy()
            if k == 0:
                B = None
            e, U = eigh(A, B)
            # e, U = eigh(A)

            atoms[k, :] = U[:, -1] / np.linalg.norm(U[:, -1])
    return atoms



import numpy as np


from .init_dict import init_dictionary
from .update_z_multi import update_z_multi
from .utils.dictionary import get_lambda_max
from .update_d_multi import update_d, update_uv

from .convolutional_dictionary_learning import DOC_FMT, DEFAULT
from .convolutional_dictionary_learning import ConvolutionalDictionaryLearning


class OnlineCDL(ConvolutionalDictionaryLearning):
    _default = {}
    _default.update(DEFAULT)
    _default['desc'] = "Online algorithm for convolutional dictionary learning"
    _default['algorithm'] = """    Online algorithm

    alpha : float
        Forgetting factor for online learning. If set to 0, the learning is
        stochastic and each D-step is independent from the previous steps.
        When set to 1, each the previous values z_hat - computed with different
        dictionary - have the same weight as the current one. This factor
        should be large enough to ensure convergence but to large factor can
        lead to sub-optimal minima.
    batch_selection : 'random' | 'cyclic'
        The batch selection strategy for online learning. The batch are either
        selected randomly among all samples (without replacement) or in a
        cyclic way.
    batch_size : int in [1, n_trials]
        Size of the batch used in online learning. Increasing it regularizes
        the dictionary learning as there is less variance for the successive
        estimates. But it also increases the computational cost as more coding
        signals z_hat must be estimate at each iteration.
    """
    __doc__ = DOC_FMT.format(**_default)

    def __init__(self, n_atoms, n_times_atom, reg=0.1, n_iter=60, n_jobs=1,
                 solver_z='lgcd', solver_z_kwargs={}, unbiased_z_hat=False,
                 solver_d='alternate_adaptive', solver_d_kwargs={},
                 rank1=True, window=False, uv_constraint='separate',
                 lmbd_max='scaled', eps=1e-10, D_init=None, D_init_params={},
                 alpha=.8, batch_size=1, batch_selection='random',
                 verbose=10, random_state=None):
        super().__init__(
            n_atoms, n_times_atom, reg=reg, n_iter=n_iter,
            solver_z=solver_z, solver_z_kwargs=solver_z_kwargs,
            rank1=rank1, window=window, uv_constraint=uv_constraint,
            unbiased_z_hat=unbiased_z_hat,
            solver_d=solver_d, solver_d_kwargs=solver_d_kwargs,
            eps=eps, D_init=D_init, D_init_params=D_init_params,
            algorithm_params=dict(alpha=alpha, batch_size=batch_size,
                                  batch_selection=batch_selection),
            n_jobs=n_jobs, random_state=random_state, algorithm='online',
            lmbd_max=lmbd_max, raise_on_increase=False, loss='l2',
            callback=None, use_sparse_z=False, verbose=verbose,
            name="OnlineCDL")

    def partial_fit(self, X, y=None):
        self._check_param_partial_fit()
        self._ensure_fit_init(X)

        # Compute the activations for the current batch and get the sufficient
        # statistic for the dictionary update
        z_hat, ztz, ztX = update_z_multi(
            X, self._D_hat, reg=self.reg_,
            solver=self.solver_z, solver_kwargs=self.solver_z_kwargs,
            loss=self.loss, loss_params=self.loss_params,
            n_jobs=self.n_jobs, return_ztz=True)

        alpha = self.algorithm_params['alpha']
        self.constants['XtX'] = X.ravel().dot(X.ravel())
        self.constants['ztz'] = alpha * self.constants['ztz'] + ztz
        self.constants['ztX'] = alpha * self.constants['ztX'] + ztX

        # Make sure the activation is not all 0
        z_nnz = np.sum(z_hat != 0, axis=(0, 2))

        if self.verbose > 5:
            print("[{}] sparsity: {:.3e}".format(
                self.name, z_nnz.sum() / z_hat.size))

        if np.all(z_nnz == 0):
            # No need to update the dictionary as this batch has all its
            # activation to 0.
            import warnings
            warnings.warn("Regularization parameter `reg` is too large and all"
                          " the activations are zero. The atoms has not been "
                          "updated.", UserWarning)
            return z_hat

        d_kwargs = dict(verbose=self.verbose, eps=1e-8)
        d_kwargs.update(self.solver_d_kwargs)
        if self.rank1:
            self._D_hat = update_uv(
                X, z_hat, uv_hat0=self._D_hat, constants=self.constants,
                solver_d=self.solver_d, uv_constraint=self.uv_constraint,
                loss=self.loss, loss_params=self.loss_params,
                window=self.window, **d_kwargs)
        else:
            self._D_hat = update_d(
                X, z_hat, D_hat0=self._D_hat, constants=self.constants,
                solver_d=self.solver_d, uv_constraint=self.uv_constraint,
                loss=self.loss, loss_params=self.loss_params,
                window=self.window, **d_kwargs)

        return z_hat

    def _ensure_fit_init(self, X):
        """Initialization for p partial_fit."""

        assert X.ndim == 3

        if hasattr(self, 'constants'):
            return

        self.constants = {}
        self.n_channels_ = n_channels = X.shape[1]

        self.constants['n_channels'] = n_channels
        self.constants['ztz'] = np.zeros((self.n_atoms, self.n_atoms,
                                          2 * self.n_times_atom - 1))
        self.constants['ztX'] = np.zeros((self.n_atoms, n_channels,
                                          self. n_times_atom))

        # Init dictionary either from D_init or from an heuristic based on the
        # first batch X
        self._D_hat = init_dictionary(
            X, self.n_atoms, self.n_times_atom, rank1=self.rank1,
            window=self.window, uv_constraint=self.uv_constraint,
            D_init=self.D_init, D_init_params=self.D_init_params,
            random_state=self.random_state)

        self.reg_ = self.reg
        _lmbd_max = get_lambda_max(X, self._D_hat).max()
        if self.lmbd_max == "scaled":
            self.reg_ *= _lmbd_max

    def _check_param_partial_fit(self):
        assert self.loss == 'l2', (
            "partial_fit is implemented only for loss={}.".format(self.loss))



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Thomas Moreau <thomas.moreau@inria.fr>
import time

import numpy as np
from scipy import optimize, sparse
from joblib import Parallel, delayed


from . import cython_code
from .utils.optim import fista
from .loss_and_gradient import gradient_zi
from .utils.lil import is_list_of_lil, is_lil
from .utils.coordinate_descent import _coordinate_descent_idx
from .utils.compute_constants import compute_DtD, compute_ztz, compute_ztX


def update_z_multi(X, D, reg, z0=None, solver='l-bfgs', solver_kwargs=dict(),
                   loss='l2', loss_params=dict(), freeze_support=False,
                   return_ztz=False, timing=False, n_jobs=1, debug=False):
    """Update z using L-BFGS with positivity constraints

    Parameters
    ----------
    X : array, shape (n_trials, n_channels, n_times)
        The data array
    D : array, shape (n_atoms, n_channels + n_times_atom)
        The dictionary used to encode the signal X. Can be either in the form
        f a full rank dictionary D (n_atoms, n_channels, n_times_atom) or with
        the spatial and temporal atoms uv (n_atoms, n_channels + n_times_atom).
    reg : float
        The regularization constant
    z0 : None | array, shape (n_trials, n_atoms, n_times_valid) |
         list of sparse lil_matrices, shape (n_atoms, n_times_valid)
        Init for z (can be used for warm restart).
    solver : 'l-bfgs' | "lgcd"
        The solver to use.
    solver_kwargs : dict
        Parameters for the solver
    loss : 'l2' | 'dtw' | 'whitening'
        The data fit loss, either classical l2 norm or the soft-DTW loss.
    loss_params : dict
        Parameters of the loss
    freeze_support : boolean
        If True, the support of z0 is frozen.
    return_ztz : boolean
        If True, returns the constants ztz and ztX, used to compute D-updates.
    timing : boolean
        If True, returns the cost function value at each iteration and the
        time taken by each iteration for each signal.
    n_jobs : int
        The number of parallel jobs.
    debug : bool
        If True, check the grad.

    Returns
    -------
    z : array, shape (n_trials, n_atoms, n_times - n_times_atom + 1)
        The true codes.
    """
    n_trials, n_channels, n_times = X.shape
    if D.ndim == 2:
        n_atoms, n_channels_n_times_atom = D.shape
        n_times_atom = n_channels_n_times_atom - n_channels
    else:
        n_atoms, n_channels, n_times_atom = D.shape
    n_times_valid = n_times - n_times_atom + 1

    if z0 is None:
        z0 = np.zeros((n_trials, n_atoms, n_times_valid))

    # now estimate the codes
    delayed_update_z = delayed(_update_z_multi_idx)

    results = Parallel(n_jobs=n_jobs)(
        delayed_update_z(X[i], D, reg, z0[i], debug, solver, solver_kwargs,
                         freeze_support, loss, loss_params=loss_params,
                         return_ztz=return_ztz, timing=timing)
        for i in np.arange(n_trials))

    # Post process the results to get separate objects
    z_hats, pobj, times = [], [], []
    if loss == 'l2' and return_ztz:
        ztz = np.zeros((n_atoms, n_atoms, 2 * n_times_atom - 1))
        ztX = np.zeros((n_atoms, n_channels, n_times_atom))
    else:
        ztz, ztX = None, None
    for z_hat, ztz_i, ztX_i, pobj_i, times_i in results:
        z_hats.append(z_hat), pobj.append(pobj_i), times.append(times_i)
        if loss == 'l2' and return_ztz:
            ztz += ztz_i
            ztX += ztX_i

    # If z_hat is a ndarray, stack and reorder the columns
    if not is_list_of_lil(z0):
        z_hats = np.array(z_hats).reshape(n_trials, n_atoms, n_times_valid)

    return z_hats, ztz, ztX


class BoundGenerator(object):
    def __init__(self, length):
        self.length = length
        self.current_index = 0

    def __len__(self):
        return self.length

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_index == self.length:
            raise StopIteration()
        self.current_index += 1
        return (0, np.inf)


def _update_z_multi_idx(X_i, D, reg, z0_i, debug, solver='l-bfgs',
                        solver_kwargs=dict(), freeze_support=False, loss='l2',
                        loss_params=dict(), return_ztz=False, timing=False):
    t_start = time.time()
    n_channels, n_times = X_i.shape
    if D.ndim == 2:
        n_atoms, n_channels_n_times_atom = D.shape
        n_times_atom = n_channels_n_times_atom - n_channels
    else:
        n_atoms, n_channels, n_times_atom = D.shape
    n_times_valid = n_times - n_times_atom + 1

    assert not (freeze_support and z0_i is None), 'Impossible !'

    if is_lil(z0_i) and solver != "lgcd":
        raise NotImplementedError()

    constants = {}
    if solver == "lgcd":
        constants['DtD'] = compute_DtD(D=D, n_channels=n_channels)
    init_timing = time.time() - t_start

    def func_and_grad(zi):
        return gradient_zi(Xi=X_i, zi=zi, D=D, constants=constants,
                           reg=reg, return_func=True, flatten=True,
                           loss=loss, loss_params=loss_params)

    if z0_i is None:
        f0 = np.zeros(n_atoms * n_times_valid)
    elif is_lil(z0_i):
        f0 = z0_i
    else:
        f0 = z0_i.reshape(n_atoms * n_times_valid)

    times, pobj = None, None
    if timing:
        times = [init_timing]
        pobj = [func_and_grad(f0)[0]]
        t_start = [time.time()]

    if solver == 'l-bfgs':
        if freeze_support:
            bounds = [(0, 0) if z == 0 else (0, None) for z in f0]
        else:
            bounds = BoundGenerator(n_atoms * n_times_valid)
        if timing:
            def callback(xk):
                times.append(time.time() - t_start[0])
                pobj.append(func_and_grad(xk)[0])
                # use a reference to have access inside this function
                t_start[0] = time.time()
        else:
            callback = None
        factr = solver_kwargs.get('factr', 1e15)  # default value
        maxiter = solver_kwargs.get('maxiter', 15000)  # default value
        z_hat, f, d = optimize.fmin_l_bfgs_b(
            func_and_grad, f0, fprime=None, args=(), approx_grad=False,
            bounds=bounds, factr=factr, maxiter=maxiter, callback=callback)

    elif solver in ("ista", "fista"):
        # Default args
        fista_kwargs = dict(
            max_iter=100, eps=None, verbose=0, restart=None,
            scipy_line_search=False,
            momentum=(solver == "fista")
        )
        fista_kwargs.update(solver_kwargs)

        def objective(z_hat):
            return func_and_grad(z_hat)[0]

        def grad(z_hat):
            return func_and_grad(z_hat)[1]

        def prox(z_hat,):
            return np.maximum(z_hat, 0.)

        output = fista(objective, grad, prox, None, f0,
                       adaptive_step_size=True, timing=timing,
                       name="Update z", **fista_kwargs)
        if timing:
            z_hat, pobj, times = output
            times[0] += init_timing
        else:
            z_hat, pobj = output

    elif solver == "lgcd":
        if not sparse.isspmatrix_lil(f0):
            f0 = f0.reshape(n_atoms, n_times_valid)

        # Default values
        tol = solver_kwargs.get('tol', 1e-1)
        n_seg = solver_kwargs.get('n_seg', 'auto')
        max_iter = solver_kwargs.get('max_iter', 1e15)
        strategy = solver_kwargs.get('strategy', 'greedy')
        output = _coordinate_descent_idx(
            X_i, D, constants, reg=reg, z0=f0,
            freeze_support=freeze_support, tol=tol, max_iter=max_iter,
            n_seg=n_seg, strategy=strategy, timing=timing, name="Update z")
        if timing:
            z_hat, pobj, times = output
            times[0] += init_timing
        else:
            z_hat = output
    else:
        raise ValueError("Unrecognized solver %s. Must be 'ista', 'fista',"
                         " 'l-bfgs', or 'lgcd'." % solver)

    if not is_lil(z_hat):
        z_hat = z_hat.reshape(n_atoms, n_times_valid)

    if loss == 'l2' and return_ztz:
        if not is_lil(z_hat):
            ztz = compute_ztz(z_hat[None], n_times_atom)
            ztX = compute_ztX(z_hat[None], X_i[None])
        else:
            cython_code._assert_cython()
            ztz = cython_code._fast_compute_ztz([z_hat], n_times_atom)
            ztX = cython_code._fast_compute_ztX([z_hat], X_i[None])
    else:
        ztz, ztX = None, None

    return z_hat, ztz, ztX, pobj, times



import numpy as np


def check_consistent_shape(*args):
    for array in args[1:]:
        if array is not None and array.shape != args[0].shape:
            raise ValueError('Incompatible shapes. Got '
                             '(%s != %s)' % (array.shape, args[0].shape))


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance.

    If seed is None, return the RandomState singleton used by np.random.
    If seed is an int, return a new RandomState instance seeded with seed.
    If seed is already a RandomState instance, return it.
    Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (int, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)



from copy import deepcopy

import numpy as np
from scipy import sparse


def convert_to_list_of_lil(z):
    return [sparse.lil_matrix(zi) for zi in z]


def convert_from_list_of_lil(z_lil):
    return np.array([zi_lil.toarray() for zi_lil in z_lil])


def get_z_shape(z):
    if is_list_of_lil(z):
        n_trials = len(z)
        n_atoms, n_times_valid = z[0].shape
    else:
        n_trials, n_atoms, n_times_valid = z.shape
    return n_trials, n_atoms, n_times_valid


def is_list_of_lil(z):
    return isinstance(z, list) and sparse.isspmatrix_lil(z[0])


def is_lil(z):
    return sparse.isspmatrix_lil(z)


def add_one_atom_in_z(z):
    n_trials, n_atoms, n_times_valid = get_z_shape(z)

    if is_list_of_lil(z):
        def add_a_zero_line(zi_lil):
            n_atoms, n_times_valid = zi_lil.shape
            new_z = sparse.lil_matrix(np.zeros((1, n_times_valid)))
            return sparse.vstack([zi_lil, new_z])

        return [add_a_zero_line(zi_lil) for zi_lil in z]
    else:
        new_z = np.zeros((n_trials, 1, n_times_valid))
        return np.concatenate([z, new_z], axis=1)


def get_nnz_and_size(z_hat):
    if is_list_of_lil(z_hat):
        z_nnz = np.array([[len(d) for d in z.data] for z in z_hat]
                         ).sum(axis=0)
        z_size = len(z_hat) * np.prod(z_hat[0].shape)
    else:
        z_nnz = np.sum(z_hat != 0, axis=(0, 2))
        z_size = z_hat.size
    return z_nnz, z_size


def init_zeros(use_sparse_z, n_trials, n_atoms, n_times_valid):
    if use_sparse_z:
        from ..cython_code import _assert_cython
        _assert_cython()
        z_hat = [sparse.lil_matrix((n_atoms, n_times_valid))
                 for _ in range(n_trials)]
    else:
        z_hat = np.zeros((n_trials, n_atoms, n_times_valid))

    return z_hat


def scale_z_by_atom(z, scale, copy=True):
    """
    Parameters
    ----------
    z_ : array, shape (n_trials, n_atoms, n_times - n_times_atom + 1)
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The sparse activation matrix.
    scale : array, shape = (n_atoms, )
        The scales to apply on z.
    """
    if is_list_of_lil(z):
        n_trials, n_atoms, n_times_valid = get_z_shape(z)
        assert n_atoms == len(scale)

        if copy:
            z = deepcopy(z)

        for z_i in z:
            for k in range(z_i.shape[0]):
                z_i.data[k] = [zikt * scale[k] for zikt in z_i.data[k]]

    else:
        if copy:
            z = z.copy()
        z *= scale[None, :, None]

    return z


def safe_sum(z, axis=None):
    n_trials, n_atoms, n_times_valid = get_z_shape(z)
    if is_list_of_lil(z):
        # n_trials = len(z) and (n_atoms, n_times_valid) = z[0].shape
        if axis is None:
            return sum([z_i.sum() for z_i in z])

        axis = list(axis)
        axis.sort()

        if axis == [0, 2]:
            res = np.zeros(n_atoms)
            for z_i in z:
                res += np.squeeze(np.array(z_i.sum(axis=1)))
            return res
        else:
            raise NotImplementedError()
    else:
        # (n_trials, n_atoms, n_times_valid) = z.shape
        return z.sum(axis=axis)



"""
Code adopted from Voytek Lab package neurodsp:
https://github.com/voytekresearch/neurodsp/blob/master/neurodsp/shape/swm.py

The sliding window matching algorithm identifies the waveform shape of
neural oscillations using correlations.
"""

# Authors: Scott Cole
#          Mainak Jas <mainak.jas@telecom-paristech.fr>

import numpy as np
from scipy.spatial.distance import pdist

from alphacsc.utils import check_random_state


def sliding_window_matching(x, L, G, max_iterations=500, T=1,
                            window_starts_custom=None, random_state=None):
    """Find recurring patterns in a time series using SWM algorithm.

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    L : float
        window length (seconds)
    G : float
        minimum window spacing (seconds)
    T : float
        temperature parameter. Controls probability of accepting a new window
    max_iterations : int
        Maximum number of iterations of potential changes in window placement
    window_starts_custom : np.ndarray (1d)
        Pre-set locations of initial windows (instead of evenly spaced by 2G)
    random_state : int
        The random state

    Returns
    -------
    avg_window : ndarray (1d)
        The average waveform in x.
    window_starts : ndarray (1d)
        Indices at which each window begins for the final set of windows
    J : np.ndarray (1d)
        Cost function value at each iteration

    References
    ----------
    Gips, B., Bahramisharif, A., Lowet, E., Roberts, M. J., de Weerd, P.,
    Jensen, O., & van der Eerden, J. (2017). Discovering recurring
    patterns in electrophysiological recordings.
    Journal of Neuroscience Methods, 275, 66-79.
    MATLAB code: https://github.com/bartgips/SWM

    Notes
    -----
    * Apply a highpass filter if looking at high frequency activity,
      so that it does not converge on a low frequency motif
    * L and G should be chosen to be about the size of the motif of interest
    """
    rng = check_random_state(random_state)

    # Initialize window positions, separated by 2*G
    if window_starts_custom is None:
        window_starts = np.arange(0, len(x) - L, 2 * G)
    else:
        window_starts = window_starts_custom
    N_windows = len(window_starts)

    # Calculate initial cost
    J = np.zeros(max_iterations)
    J[0] = _compute_J(x, window_starts, L)

    # Randomly sample windows with replacement
    random_window_idx = rng.choice(range(N_windows), size=max_iterations)

    # For each iteration, randomly replace a window with a new window
    # to improve cross-window similarity
    for idx in range(1, max_iterations):
        # Pick a random window position
        window_idx_replace = random_window_idx[idx]

        # Find a new allowed position for the window
        window_starts_temp = np.copy(window_starts)
        window_starts_temp[window_idx_replace] = _find_new_windowidx(
            window_starts, G, L, len(x) - L, rng)

        # Calculate the cost with replaced windows
        J_temp = _compute_J(x, window_starts_temp, L)

        # Calculate the change in cost function
        deltaJ = J_temp - J[idx - 1]

        # Calculate the acceptance probability
        p_accept = np.exp(-deltaJ / float(T))

        # Accept update to J with a certain probability
        if rng.rand() < p_accept:
            J[idx] = J_temp
            # Update X
            window_starts = window_starts_temp
        else:
            J[idx] = J[idx - 1]

        print('[iter %03d] Cost function: %s' % (idx, J[idx]))

    # Calculate average window
    avg_window = np.zeros(L)
    for w in range(N_windows):
        avg_window += x[window_starts[w]:window_starts[w] + L]
    avg_window = avg_window / float(N_windows)

    return avg_window, window_starts, J


def _compute_J(x, window_starts, L):
    """Compute the cost, which is proportional to the
    difference between pairs of windows"""

    # Get all windows and zscore them
    N_windows = len(window_starts)
    windows = np.zeros((N_windows, L))
    for w in range(N_windows):
        temp = x[window_starts[w]:window_starts[w] + L]
        windows[w] = (temp - np.mean(temp)) / np.std(temp)

    # Calculate distances for all pairs of windows
    dist = pdist(np.vstack(windows),
                 lambda u, v: np.sum((u - v) ** 2))
    J = np.sum(dist) / float(L * (N_windows - 1))

    return J


def _find_new_windowidx(window_starts, G, L, N_samp, rng,
                        tries_limit=1000):
    """Find a new sample for the starting window"""

    found = False
    N_tries = 0
    while found is False:
        # Generate a random sample
        new_samp = rng.randint(N_samp)
        # Check how close the sample is to other window starts
        dists = np.abs(window_starts - new_samp)
        if np.min(dists) > G:
            return new_samp
        else:
            N_tries += 1
            if N_tries > tries_limit:
                raise RuntimeError('SWM algorithm has difficulty finding a new'
                                   ' window. Increase the spacing parameter,'
                                   ' G.')



import numpy as np
from numpy.fft import rfft, irfft
from scipy.fftpack import next_fast_len  # noqa

from sklearn.metrics.pairwise import check_pairwise_arrays


def roll_invariant_euclidean_distances(X, Y=None, squared=False):
    """
    Considering the rows of X (and Y=X) as vectors, compute the
    distance matrix between each pair of vectors.
    The distance is the minimum of the euclidean distance over all rolls:

        dist(x, y) = min_\tau(||x(t) - y(t - \tau)||^2)

    Parameters
    ----------
    X : array, shape (n_samples_1, n_features)

    Y : array, shape (n_samples_2, n_features)

    squared : boolean
        Not used. Only for API compatibility.

    Returns
    -------
    distances : array, shape (n_samples_1, n_samples_2)

    """
    X = np.atleast_2d(X)
    if Y is not None:
        Y = np.atleast_2d(Y)
    X, Y = check_pairwise_arrays(X, Y)
    n_samples_1, n_features = X.shape
    n_samples_2, n_features = Y.shape

    X_norm = np.power(np.linalg.norm(X, axis=1), 2)
    Y_norm = np.power(np.linalg.norm(Y, axis=1), 2)

    # n_pads = 0
    # n_fft = next_fast_len(n_features + n_pads)
    n_fft = n_features  # not fast but otherwise the distance is wrong
    X_hat = rfft(X, n_fft, axis=1)
    Y_hat = rfft(Y, n_fft, axis=1).conj()

    # # broadcasting can have a huge memory cost
    # XY_hat = X_hat[:, None, :] * Y_hat[None, :, :]
    # XY = irfft(XY_hat, n_fft, axis=2).max(axis=2)
    # distances = X_norm[:, None] + Y_norm[None, :] - 2 * XY

    distances = np.zeros((n_samples_1, n_samples_2))
    if n_samples_2 > 1:
        print('RIED on %s samples, this might be slow' % (distances.shape, ))
    for ii in range(n_samples_1):
        for jj in range(n_samples_2):
            XY = irfft(X_hat[ii] * Y_hat[jj], n_fft).max()
            distances[ii, jj] = X_norm[ii] + Y_norm[jj] - 2 * XY

    distances += 1e-12

    return distances


def translation_invariant_euclidean_distances(X, Y=None, squared=False,
                                              symmetric=False):
    """
    Considering the rows of X (and Y=X) as vectors, compute the
    distance matrix between each pair of vectors.
    The distance is the minimum of the euclidean distance over a set of
    translations:

        dist(x, y) = min_{i, j}(||x(i:i+T) - y(j:j+T)||^2)

    where T = n_features / 2, and 1 <= i, j <= n_features / 2

    Parameters
    ----------
    X : array, shape (n_samples_1, n_features)

    Y : array, shape (n_samples_2, n_features)

    squared : boolean
        Not used. Only for API compatibility.

    symmetric : boolean
        If False, the distance is not symmetric anymore, since we keep indice
        j fixed at `n_features / 4`.

    Returns
    -------
    distances : array, shape (n_samples_1, n_samples_2)

    """
    X = np.atleast_2d(X)
    if Y is not None:
        Y = np.atleast_2d(Y)
    X, Y = check_pairwise_arrays(X, Y)
    n_samples_1, n_features = X.shape
    n_samples_2, n_features = Y.shape

    distances = np.zeros((n_samples_1, n_samples_2))
    # if n_samples_2 > 1:
    #     print('TIED on %s samples, this might be slow' % (distances.shape, ))
    for nn in range(n_samples_1):
        for mm in range(n_samples_2):
            XY = (X[nn, :, None] - Y[mm, None, :]) ** 2

            if symmetric:
                jj_range = np.arange(n_features // 2)
            else:
                jj_range = [n_features // 4]

            dist = np.zeros((n_features // 2, len(jj_range)))
            for ii in range(n_features // 2):
                for jj, kk in enumerate(jj_range):
                    xy = XY[ii:ii + n_features // 2, kk:kk + n_features // 2]
                    dist[ii, jj] = xy.trace(axis1=0, axis2=1)
            distances[nn, mm] = dist.min()

    return distances



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>

import numpy as np

from .utils import construct_X, check_random_state
from .learn_d_z import learn_d_z
from .update_d import update_d_block
from .update_w import estimate_phi_mh


def learn_d_z_weighted(
        X, n_atoms, n_times_atom, func_d=update_d_block, reg=0.1, alpha=1.9,
        lmbd_max='fixed', n_iter_global=10, init_tau=False, n_iter_optim=10,
        n_iter_mcmc=10, n_burnin_mcmc=0, random_state=None, n_jobs=1,
        solver_z='l-bfgs', solver_d_kwargs=dict(), solver_z_kwargs=dict(),
        ds_init=None, ds_init_params=dict(), verbose=0, callback=None):
    """Univariate Convolutional Sparse Coding with an alpha-stable distribution

    Parameters
    ----------
    X : array, shape (n_trials, n_times)
        The data on which to perform CSC.
    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    func_d : callable
        The function to update the atoms.
    reg : float
        The regularization parameter
    alpha : float in [0, 2[:
        Parameter of the alpha-stable noise distribution.
    lmbd_max : 'fixed' | 'scaled' | 'per_atom' | 'shared'
        If not fixed, adapt the regularization rate as a ratio of lambda_max:
          - 'scaled': the regularization parameter is fixed as a ratio of its
            maximal value at init i.e. reg_used = reg * lmbd_max(uv_init)
          - 'shared': the regularization parameter is set at each iteration as
            a ratio of its maximal value for the current dictionary estimate
            i.e. reg_used = reg * lmbd_max(uv_hat)
          - 'per_atom': the regularization parameter is set per atom and at
            each iteration as a ratio of its maximal value for this atom i.e.
            reg_used[k] = reg * lmbd_max(uv_hat[k])
    n_iter_global : int
        The number of iteration of the Expectation-Maximisation outer loop.
    init_tau : boolean
        If True, use a heuristic to initialize the weights tau.
    n_iter_optim : int
        The number of iteration of the Maximisation step (weighted CSC).
    n_iter_mcmc : int
        The number of iteration of the Expectation step (MCMC).
    n_burnin_mcmc : int
        The number of iteration unused by the MCMC algorithm.
    random_state : int | None
        The random state.
    n_jobs : int
        The number of parallel jobs.
    solver_z : str
        The solver to use for the z update. Options are
        'l-bfgs' (default) | 'ista' | 'fista'
    solver_d_kwargs : dict
        Additional keyword arguments to provide to update_d
    solver_z_kwargs : dict
        Additional keyword arguments to pass to update_z
    ds_init_params : dict
        Dictionnary of parameters for the kmeans init method.
    ds_init_params : dict
        Dictionnary of parameters for the kmeans init method.
    verbose : int
        The verbosity level.
    callback : func
        A callback function called at the end of each loop of the
        coordinate descent.

    Returns
    -------
    d_hat : array, shape (n_atoms, n_times_atom)
        The estimated atoms.
    z_hat : array, shape (n_atoms, n_trials, n_times - n_times_atom + 1)
        The sparse activation matrix.
    tau : array, shape (n_trials, n_times)
        Weights estimated by the Expectation-Maximisation algorithm.
    """

    n_trials, n_times = X.shape

    if init_tau:
        phi = np.tile(np.var(X, axis=1)[:, None], X.shape[1])
        tau = 1 / phi
    else:
        # assume gaussian to start with
        phi = np.full(shape=(n_trials, n_times), fill_value=2.0)
        tau = np.full(shape=(n_trials, n_times), fill_value=0.5)

    rng = check_random_state(random_state)
    d_hat = ds_init
    z_hat = None
    # Run the MCEM algorithm
    for ii in range(n_iter_global):

        if verbose > 0:
            print("Global Iter: %d/%d\t" % (ii, n_iter_global), end='',
                  flush=True)

        # E-step: Estimate the expectation via MCMC
        if ii == 0:
            X_hat = np.zeros_like(X)
        else:
            X_hat = construct_X(z_hat, d_hat)
        phi, tau, loglk_mcmc = estimate_phi_mh(
            X, X_hat, alpha, phi, n_iter_mcmc, n_burnin_mcmc, random_state=rng,
            return_loglk=True, verbose=verbose)

        # M-step: Optimize d and z wrt the new weights
        pobj, times, d_hat, z_hat, reg = learn_d_z(
            X, n_atoms, n_times_atom, func_d, reg=reg, lmbd_max=lmbd_max,
            n_iter=n_iter_optim, random_state=rng, sample_weights=2 * tau,
            ds_init=d_hat, ds_init_params=ds_init_params,
            solver_d_kwargs=solver_d_kwargs, solver_z_kwargs=solver_z_kwargs,
            verbose=verbose, solver_z=solver_z, n_jobs=n_jobs,
            callback=callback)
        lmbd_max = 'fixed'  # subsequent iterations use the same regularization

    return d_hat, z_hat, tau



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Thomas Moreau <thomas.moreau@inria.fr>

from __future__ import print_function
import time
import sys

import numpy as np

from .utils import lil
from .utils import check_random_state
from .utils.convolution import sort_atoms_by_explained_variances
from .utils.dictionary import get_lambda_max
from .utils.whitening import whitening
from .init_dict import init_dictionary, get_max_error_dict
from .loss_and_gradient import compute_X_and_objective_multi
from .update_z_multi import update_z_multi
from .update_d_multi import update_uv, update_d


def learn_d_z_multi(X, n_atoms, n_times_atom, n_iter=60, n_jobs=1,
                    lmbd_max='fixed', reg=0.1, loss='l2',
                    loss_params=dict(gamma=.1, sakoe_chiba_band=10, ordar=10),
                    rank1=True, uv_constraint='separate', eps=1e-10,
                    algorithm='batch', algorithm_params=dict(),
                    detrending=None, detrending_params=dict(),
                    solver_z='l-bfgs', solver_z_kwargs=dict(),
                    solver_d='alternate_adaptive', solver_d_kwargs=dict(),
                    D_init=None, D_init_params=dict(),
                    unbiased_z_hat=False, use_sparse_z=False,
                    stopping_pobj=None, raise_on_increase=True,
                    verbose=10, callback=None, random_state=None, name="DL",
                    window=False, sort_atoms=False):
    """Multivariate Convolutional Sparse Coding with optional rank-1 constraint

    Parameters
    ----------
    X : array, shape (n_trials, n_channels, n_times)
        The data on which to perform CSC.
    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    reg : float
        The regularization parameter
    lmbd_max : 'fixed' | 'scaled' | 'per_atom' | 'shared'
        If not fixed, adapt the regularization rate as a ratio of lambda_max:
          - 'scaled': the regularization parameter is fixed as a ratio of its
            maximal value at init i.e. reg_used = reg * lmbd_max(uv_init)
          - 'shared': the regularization parameter is set at each iteration as
            a ratio of its maximal value for the current dictionary estimate
            i.e. reg_used = reg * lmbd_max(uv_hat)
          - 'per_atom': the regularization parameter is set per atom and at
            each iteration as a ratio of its maximal value for this atom i.e.
            reg_used[k] = reg * lmbd_max(uv_hat[k])
    n_iter : int
        The number of coordinate-descent iterations.
    n_jobs : int
        The number of parallel jobs.
    loss : 'l2' | 'dtw'
        Loss for the data-fit term. Either the norm l2 or the soft-DTW.
    loss_params : dict
        Parameters of the loss
    rank1 : boolean
        If set to True, learn rank 1 dictionary atoms.
    uv_constraint : str in {'joint' | 'separate'}
        The kind of norm constraint on the atoms:
        If 'joint', the constraint is norm_2([u, v]) <= 1
        If 'separate', the constraint is norm_2(u) <= 1 and norm_2(v) <= 1
    eps : float
        Stopping criterion. If the cost descent after a uv and a z update is
        smaller than eps, return.
    algorithm : 'batch' | 'greedy' | 'online'
        Dictionary learning algorithm.
    algorithm_params : dict
        Parameters for the global algorithm used to learn the dictionary:
          alpha : float
            Forgetting factor for online learning. If set to 0, the learning is
            stochastic and each D-step is independent from the previous steps.
            When set to 1, each the previous values z_hat - computed with
            different dictionary - have the same weight as the current one.
            This factor should be large enough to ensure convergence but to
            large factor can lead to sub-optimal minima.
          batch_selection : 'random' | 'cyclic'
            The batch selection strategy for online learning. The batch are
            either selected randomly among all samples (without replacement) or
            in a cyclic way.
          batch_size : int in [1, n_trials]
            Size of the batch used in online learning. Increasing it
            regularizes the dictionary learning as there is less variance for
            the successive estimates. But it also increases the computational
            cost as more coding signals z_hat must be estimate at each
            iteration.
    solver_z : str
        The solver to use for the z update. Options are
        'l-bfgs' (default) | "lgcd"
    solver_z_kwargs : dict
        Additional keyword arguments to pass to update_z_multi
    solver_d : str
        The solver to use for the d update. Options are
        'alternate' | 'alternate_adaptive' (default) | 'joint'
    solver_d_kwargs : dict
        Additional keyword arguments to provide to update_d
    D_init : str or array, shape (n_atoms, n_channels + n_times_atoms) or \
                           shape (n_atoms, n_channels, n_times_atom)
        The initial atoms or an initialization scheme in {'kmeans' | 'ssa' |
        'chunks' | 'random'}.
    D_init_params : dict
        Dictionnary of parameters for the kmeans init method.
    unbiased_z_hat : boolean
        If set to True, the value of the non-zero coefficients in the returned
        z_hat are recomputed with reg=0 on the frozen support.
    use_sparse_z : boolean
        Use sparse lil_matrices to store the activations.
    verbose : int
        The verbosity level.
    callback : func
        A callback function called at the end of each loop of the
        coordinate descent.
    random_state : int | None
        The random state.
    raise_on_increase : boolean
        Raise an error if the objective function increase
    window : boolean
        If True, re-parametrizes the atoms with a temporal Tukey window
    sort_atoms : boolean
        If True, the atoms are sorted by explained variances.

    Returns
    -------
    pobj : list
        The objective function value at each step of the coordinate descent.
    times : list
        The cumulative time for each iteration of the coordinate descent.
    uv_hat : array, shape (n_atoms, n_channels + n_times_atom)
        The atoms to learn from the data.
    z_hat : array, shape (n_trials, n_atoms, n_times_valid)
        The sparse activation matrix.
    reg : float
        Regularization parameter used.
    """

    assert lmbd_max in ['fixed', 'scaled', 'per_atom', 'shared'], (
        "lmbd_max should be in {'fixed', 'scaled', 'per_atom', 'shared'}, "
        "not '{}'".format(lmbd_max)
    )

    n_trials, n_channels, n_times = X.shape
    n_times_valid = n_times - n_times_atom + 1

    # initialization
    start = time.time()
    rng = check_random_state(random_state)

    D_hat = init_dictionary(X, n_atoms, n_times_atom, D_init=D_init,
                            rank1=rank1, uv_constraint=uv_constraint,
                            D_init_params=D_init_params, random_state=rng,
                            window=window)
    b_hat_0 = rng.randn(n_atoms * (n_channels + n_times_atom))
    init_duration = time.time() - start

    z_hat = lil.init_zeros(use_sparse_z, n_trials, n_atoms, n_times_valid)

    z_kwargs = dict(verbose=verbose)
    z_kwargs.update(solver_z_kwargs)

    # Compute the coefficients to whiten X. TODO: add timing
    if loss == 'whitening':
        loss_params['ar_model'], X = whitening(X, ordar=loss_params['ordar'])

    _lmbd_max = get_lambda_max(X, D_hat).max()
    if verbose > 1:
        print("[{}] Max value for lambda: {}".format(name, _lmbd_max))
    if lmbd_max == "scaled":
        reg = reg * _lmbd_max

    def compute_z_func(X, z_hat, D_hat, reg=None):
        return update_z_multi(X, D_hat, reg=reg, z0=z_hat,
                              solver=solver_z, solver_kwargs=z_kwargs,
                              loss=loss, loss_params=loss_params,
                              n_jobs=n_jobs, return_ztz=True)

    def obj_func(X, z_hat, D_hat, reg=None, return_X_hat=False):
        return compute_X_and_objective_multi(X, z_hat, D_hat,
                                             reg=reg, loss=loss,
                                             loss_params=loss_params,
                                             uv_constraint=uv_constraint,
                                             feasible_evaluation=True,
                                             return_X_hat=return_X_hat)

    d_kwargs = dict(verbose=verbose, eps=1e-8)
    d_kwargs.update(solver_d_kwargs)
    if algorithm == "stochastic":
        # The typical stochastic algorithm samples one signal, compute the
        # associated value z and then perform one step of gradient descent
        # for D.
        d_kwargs["max_iter"] = 1

    def compute_d_func(X, z_hat, D_hat, constants):
        if rank1:
            return update_uv(X, z_hat, uv_hat0=D_hat, constants=constants,
                             b_hat_0=b_hat_0, solver_d=solver_d,
                             uv_constraint=uv_constraint, loss=loss,
                             loss_params=loss_params, window=window,
                             **d_kwargs)
        else:
            return update_d(X, z_hat, D_hat0=D_hat, constants=constants,
                            b_hat_0=b_hat_0, solver_d=solver_d,
                            uv_constraint=uv_constraint, loss=loss,
                            loss_params=loss_params, window=window, **d_kwargs)

    if callable(callback):
        callback(X, D_hat, z_hat, [])

    end_iter_func = get_iteration_func(eps, stopping_pobj, callback, lmbd_max,
                                       name, verbose, raise_on_increase)

    # common parameters
    kwargs = dict(X=X, D_hat=D_hat, z_hat=z_hat, compute_z_func=compute_z_func,
                  compute_d_func=compute_d_func, obj_func=obj_func,
                  end_iter_func=end_iter_func, n_iter=n_iter, verbose=verbose,
                  random_state=random_state, window=window, reg=reg,
                  lmbd_max=lmbd_max, name=name, uv_constraint=uv_constraint)
    kwargs.update(algorithm_params)

    if algorithm == 'batch':
        pobj, times, D_hat, z_hat = _batch_learn(greedy=False, **kwargs)
    elif algorithm == "greedy":
        pobj, times, D_hat, z_hat = _batch_learn(greedy=True, **kwargs)
    elif algorithm == "online":
        pobj, times, D_hat, z_hat = _online_learn(**kwargs)
    elif algorithm == "stochastic":
        # For stochastic learning, set forgetting factor alpha of the
        # online algorithm to 0, making each step independent of previous
        # steps and set D-update max_iter to a low value (typically 1).
        kwargs['alpha'] = 0
        pobj, times, D_hat, z_hat = _online_learn(**kwargs)
    else:
        raise NotImplementedError(
            "Algorithm '{}' is not implemented to learn dictionary atoms."
            .format(algorithm))

    if sort_atoms:
        D_hat, z_hat = sort_atoms_by_explained_variances(
            D_hat, z_hat, n_channels=n_channels)

    # recompute z_hat with no regularization and keeping the support fixed
    if unbiased_z_hat:
        start_unbiased_z_hat = time.time()
        z_hat, _, _ = update_z_multi(
            X, D_hat, reg=0, z0=z_hat, n_jobs=n_jobs, solver=solver_z,
            solver_kwargs=solver_z_kwargs, freeze_support=True, loss=loss,
            loss_params=loss_params)
        if verbose > 1:
            print("[{}] Compute the final z_hat with support freeze in "
                  "{:.2f}s".format(name, time.time() - start_unbiased_z_hat))

    times[0] += init_duration

    if verbose > 0:
        print("[%s] Fit in %.1fs" % (name, time.time() - start))

    return pobj, times, D_hat, z_hat, reg


def _batch_learn(X, D_hat, z_hat, compute_z_func, compute_d_func,
                 obj_func, end_iter_func, n_iter=100,
                 lmbd_max='fixed', reg=None, verbose=0, greedy=False,
                 random_state=None, name="batch", uv_constraint='separate',
                 window=False):

    reg_ = reg

    # Initialize constants dictionary
    constants = {}
    constants['n_channels'] = X.shape[1]
    constants['XtX'] = np.dot(X.ravel(), X.ravel())

    if greedy:
        n_iter_by_atom = 1
        # remove all atoms
        D_hat = D_hat[0:0]
        # remove all activations
        use_sparse_z = lil.is_list_of_lil(z_hat)
        n_trials, n_atoms, n_times_valid = lil.get_z_shape(z_hat)
        z_hat = lil.init_zeros(use_sparse_z, n_trials, 0, n_times_valid)

        if n_iter < n_atoms:
            raise ValueError('The greedy method needs at least %d iteration '
                             'to learn %d atoms. Got only n_iter=%d. Please '
                             'increase n_iter.' % (
                                 n_iter_by_atom * n_atoms, n_atoms, n_iter))

    # monitor cost function
    times = [0]
    pobj = [obj_func(X, z_hat, D_hat, reg=reg_)]

    for ii in range(n_iter):  # outer loop of coordinate descent
        if verbose == 1:
            msg = '.' if ((ii + 1) % 50 != 0) else '+\n'
            print(msg, end='')
            sys.stdout.flush()
        if verbose > 1:
            print('[{}] CD iterations {} / {}'.format(name, ii, n_iter))

        if greedy and ii % n_iter_by_atom == 0 and D_hat.shape[0] < n_atoms:
            # add a new atom every n_iter_by_atom iterations
            new_atom = get_max_error_dict(X, z_hat, D_hat,
                                          uv_constraint=uv_constraint,
                                          window=window)[0]
            D_hat = np.concatenate([D_hat, new_atom[None]])
            z_hat = lil.add_one_atom_in_z(z_hat)

        if lmbd_max not in ['fixed', 'scaled']:
            reg_ = reg * get_lambda_max(X, D_hat)
            if lmbd_max == 'shared':
                reg_ = reg_.max()

        if verbose > 5:
            print('[{}] lambda = {:.3e}'.format(name, np.mean(reg_)))

        # Compute z update
        start = time.time()
        z_hat, constants['ztz'], constants['ztX'] = compute_z_func(
            X, z_hat, D_hat, reg=reg_)

        # monitor cost function
        times.append(time.time() - start)
        pobj.append(obj_func(X, z_hat, D_hat, reg=reg_))

        z_nnz, z_size = lil.get_nnz_and_size(z_hat)
        if verbose > 5:
            print("[{}] sparsity: {:.3e}".format(
                name, z_nnz.sum() / z_size))
            print('[{}] Objective (z) : {:.3e}'.format(name, pobj[-1]))

        if np.all(z_nnz == 0):
            import warnings
            warnings.warn("Regularization parameter `reg` is too large "
                          "and all the activations are zero. No atoms has"
                          " been learned.", UserWarning)
            break

        # Compute D update
        start = time.time()
        D_hat = compute_d_func(X, z_hat, D_hat, constants)

        # monitor cost function
        times.append(time.time() - start)
        pobj.append(obj_func(X, z_hat, D_hat, reg=reg_))

        null_atom_indices = np.where(z_nnz == 0)[0]
        if len(null_atom_indices) > 0:
            k0 = null_atom_indices[0]
            D_hat[k0] = get_max_error_dict(X, z_hat, D_hat, uv_constraint,
                                           window=window)[0]
            if verbose > 5:
                print('[{}] Resampled atom {}'.format(name, k0))

        if verbose > 5:
            print('[{}] Objective (d) : {:.3e}'.format(name, pobj[-1]))

        if end_iter_func(X, z_hat, D_hat, pobj, ii):
            break

    return pobj, times, D_hat, z_hat


def _online_learn(X, D_hat, z_hat, compute_z_func, compute_d_func,
                  obj_func, end_iter_func, n_iter=100, verbose=0,
                  random_state=None, lmbd_max='fixed', reg=None,
                  alpha=.8, batch_selection='random', batch_size=1,
                  name="online", uv_constraint='separate', window=False):

    reg_ = reg

    # Initialize constants dictionary
    constants = {}
    n_trials, n_channels = X.shape[:2]
    if D_hat.ndim == 2:
        n_atoms, n_times_atom = D_hat.shape
        n_times_atom -= n_channels
    else:
        n_atoms, _, n_times_atom = D_hat.shape
    constants['n_channels'] = n_channels
    constants['XtX'] = np.dot(X.ravel(), X.ravel())
    constants['ztz'] = np.zeros((n_atoms, n_atoms, 2 * n_times_atom - 1))
    constants['ztX'] = np.zeros((n_atoms, n_channels, n_times_atom))

    # monitor cost function
    times = [0]
    pobj = [obj_func(X, z_hat, D_hat, reg=reg_)]

    for ii in range(n_iter):  # outer loop of coordinate descent
        if verbose == 1:
            msg = '.' if (ii % 50 != 0) else '+\n'
            print(msg, end='')
            sys.stdout.flush()
        if verbose > 1:
            print('[{}] CD iterations {} / {}'.format(name, ii, n_iter))

        if lmbd_max not in ['fixed', 'scaled']:
            reg_ = reg * get_lambda_max(X, D_hat)
            if lmbd_max == 'shared':
                reg_ = reg_.max()

        if verbose > 5:
            print('[{}] lambda = {:.3e}'.format(name, np.mean(reg_)))

        # Compute z update
        start = time.time()
        if batch_selection == 'random':
            i0 = np.random.choice(n_trials, batch_size, replace=False)
        elif batch_selection == 'cyclic':
            i_slice = (ii * batch_size) % n_trials
            i0 = slice(i_slice, i_slice + batch_size)
        else:
            raise NotImplementedError(
                "the '{}' batch_selection strategy for the online learning is "
                "not implemented.".format(batch_selection))
        z_hat[i0], ztz, ztX = compute_z_func(X[i0], z_hat[i0], D_hat, reg=reg_)

        constants['ztz'] = alpha * constants['ztz'] + ztz
        constants['ztX'] = alpha * constants['ztX'] + ztX

        # monitor cost function
        times.append(time.time() - start)
        pobj.append(obj_func(X, z_hat, D_hat, reg=reg_))

        z_nnz, z_size = lil.get_nnz_and_size(z_hat)
        if verbose > 5:
            print("[{}] sparsity: {:.3e}".format(
                name, z_nnz.sum() / z_size))
            print('[{}] Objective (z) : {:.3e}'.format(name, pobj[-1]))

        if np.all(z_nnz == 0):
            import warnings
            warnings.warn("Regularization parameter `reg` is too large "
                          "and all the activations are zero. No atoms has"
                          " been learned.", UserWarning)
            break

        # Compute D update
        start = time.time()
        D_hat = compute_d_func(X, z_hat, D_hat, constants)

        # monitor cost function
        times.append(time.time() - start)
        pobj.append(obj_func(X, z_hat, D_hat, reg=reg_))

        null_atom_indices = np.where(z_nnz == 0)[0]
        if len(null_atom_indices) > 0:
            k0 = null_atom_indices[0]
            D_hat[k0] = get_max_error_dict(X, z_hat, D_hat, uv_constraint,
                                           window=window)[0]
            if verbose > 5:
                print('[{}] Resampled atom {}'.format(name, k0))

        if verbose > 5:
            print('[{}] Objective (d) : {:.3e}'.format(name, pobj[-1]))

        if end_iter_func(X, z_hat, D_hat, pobj, ii):
            break

    return pobj, times, D_hat, z_hat


def get_iteration_func(eps, stopping_pobj, callback, lmbd_max, name, verbose,
                       raise_on_increase):
    def end_iteration(X, z_hat, D_hat, pobj, iteration):
        if callable(callback):
            callback(X, D_hat, z_hat, pobj)

        # Only check that the cost is always going down when the regularization
        # parameter is fixed.
        dz = (pobj[-3] - pobj[-2]) / min(pobj[-3], pobj[-2])
        du = (pobj[-2] - pobj[-1]) / min(pobj[-2], pobj[-1])
        if ((dz < eps or du < eps) and lmbd_max in ['fixed', 'scaled']):
            if dz < 0 and raise_on_increase:
                raise RuntimeError(
                    "The z update have increased the objective value by {}."
                    .format(dz))
            if du < -1e-10 and dz > 1e-12 and raise_on_increase:
                raise RuntimeError(
                    "The d update have increased the objective value by {}."
                    "(dz={})".format(du, dz))
            if dz < eps and du < eps:
                if verbose == 1:
                    print("")
                print("[{}] Converged after {} iteration, (dz, du) "
                      "= {:.3e}, {:.3e}".format(name, iteration + 1, dz, du))
                return True

        if stopping_pobj is not None and pobj[-1] < stopping_pobj:
            return True
        return False

    return end_iteration



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>

import numpy as np
from scipy import linalg, optimize

from .utils import construct_X, check_consistent_shape


def update_d(X, Z, n_times_atom, lambd0=None, ds_init=None, debug=False,
             solver_kwargs=dict(), sample_weights=None, verbose=0):
    """Learn d's in time domain.

    Parameters
    ----------
    X : array, shape (n_trials, n_times)
        The data for sparse coding
    Z : array, shape (n_atoms, n_trials, n_times - n_times_atom + 1)
        The code for which to learn the atoms
    n_times_atom : int
        The shape of atoms.
    lambd0 : array, shape (n_atoms,) | None
        The init for lambda.
    debug : bool
        If True, check grad.
    solver_kwargs : dict
        Parameters for the solver
    sample_weights: array, shape (n_trials, n_times)
        Weights applied on the cost function.
    verbose : int
        Verbosity level.

    Returns
    -------
    d_hat : array, shape (k, n_times_atom)
        The atom to learn from the data.
    lambd_hats : float
        The dual variables
    """
    check_consistent_shape(X, sample_weights)
    n_atoms, n_trials, n_times_valid = Z.shape
    n_times = n_times_valid + n_times_atom - 1

    if lambd0 is None:
        lambd0 = 10. * np.ones(n_atoms)

    lhs = np.zeros((n_times_atom * n_atoms, ) * 2)
    rhs = np.zeros(n_times_atom * n_atoms)
    Zki = np.zeros(n_times + (n_times_atom - 1))
    for i in range(n_trials):

        ZZi = []
        for k in range(n_atoms):
            Zki[n_times_atom - 1:-(n_times_atom - 1)] = Z[k, i]
            ZZik = _embed(Zki, n_times_atom)
            # n_times_atom, n_times = ZZik.shape
            ZZi.append(ZZik)

        ZZi = np.concatenate(ZZi, axis=0)
        # n_times_atom * n_atoms, n_times = ZZi.shape
        if sample_weights is not None:
            wZZi = sample_weights[i] * ZZi
        else:
            wZZi = ZZi
        lhs += np.dot(wZZi, ZZi.T)
        rhs += np.dot(wZZi, X[i])

    factr = solver_kwargs.get('factr', 1e7)  # default value
    d_hat, lambd_hats = solve_unit_norm_dual(lhs, rhs, lambd0=lambd0,
                                             factr=factr, debug=debug,
                                             lhs_is_toeplitz=False)
    d_hat = d_hat.reshape(n_atoms, n_times_atom)[:, ::-1]
    return d_hat, lambd_hats


def update_d_block(X, Z, n_times_atom, lambd0=None, ds_init=None,
                   projection='dual', n_iter=1, solver_kwargs=dict(),
                   sample_weights=None, verbose=0):
    """Learn d's in time domain but each atom is learned separately.

    Parameters
    ----------
    X : array, shape (n_trials, n_times)
        The data for sparse coding
    Z : array, shape (n_atoms, n_trials, n_times - n_times_atom + 1)
        The code from which to learn the atoms
    n_times_atom : int
        The size of atoms.
    lambd0 : array, shape (n_atoms,) | None
        The init for dual variables
    ds_init : array, shape (n_atoms, n_times_atom)
        Warm start d_hats for the l-bfgs / projected gradient solver.
    projection : str, ('primal', 'dual')
        Whether to project to unit ball in the primal or the dual.
    n_iter : int
        Number of outer loop over atoms
    solver_kwargs : dict
        Parameters for the solver
    sample_weights: array, shape (n_trials, n_times)
        Weights applied on the cost function.
    verbose : int
        Verbosity level.

    Returns
    -------
    d_hats : array, shape (n_atoms, n_times_atom)
        The atom to learn from the data.
    lambd_hats : float
        The dual variables
    """
    check_consistent_shape(X, sample_weights)
    n_atoms, n_trials, n_times_valid = Z.shape
    n_times = n_times_valid + n_times_atom - 1

    if lambd0 is None:
        lambd0 = 10. * np.ones(n_atoms)

    Zki = np.zeros(n_times + n_times_atom - 1)

    if ds_init is None:
        ds = np.zeros((n_atoms, n_times_atom))
    else:
        ds = ds_init.copy()

    lambd_hats = np.zeros(n_atoms)
    residual = X - construct_X(Z, ds)
    if verbose > 1:
        print('Using method %s for projection' % projection)
    for _ in range(n_iter):
        for k in range(n_atoms):
            residual += construct_X(Z[k:k + 1], ds[k:k + 1])

            if sample_weights is not None:
                # with sample_weights, lhs is not toeplitz
                lhs = np.zeros((n_times_atom, n_times_atom))
            else:
                # lhs_c is only the first column of the full toeplitz matrix
                lhs_c = np.zeros(n_times_atom)

            rhs = np.zeros(n_times_atom)
            for i in range(n_trials):
                Zki[n_times_atom - 1:-(n_times_atom - 1)] = Z[k, i]
                ZZi = _embed(Zki, n_times_atom).copy()
                # n_times_atom, n_times = ZZik.shape

                if sample_weights is not None:
                    wZZi = sample_weights[i] * ZZi
                    lhs += np.dot(wZZi, ZZi.T)
                    rhs += np.dot(wZZi, residual[i])
                else:
                    lhs_c += np.dot(ZZi[0], ZZi.T)
                    rhs += np.dot(ZZi, residual[i])

            if sample_weights is None:
                # transforming the column into a squared toeplitz matrix
                lhs = linalg.toeplitz(lhs_c)

            if projection == 'primal':
                factr = solver_kwargs.get('factr', 1e11)  # default value
                d_hat = solve_unit_norm_primal(lhs, rhs, d_hat0=ds[k][::-1],
                                               factr=factr, verbose=verbose)

            elif projection == 'dual':
                factr = solver_kwargs.get('factr', 1e7)  # default value
                d_hat, lambd_hat = solve_unit_norm_dual(
                    lhs, rhs, factr=factr, lambd0=np.array([lambd0[k]]),
                    lhs_is_toeplitz=sample_weights is None)

                lambd_hats[k] = lambd_hat
            else:
                raise ValueError('Unknown projection %s.' % projection)
            ds[k] = d_hat[::-1]  # reversal for convolution
            residual -= construct_X(Z[k:k + 1], ds[k:k + 1])
    return ds, lambd_hats


def _embed(x, dim, lag=1):
    """Create an embedding of array given a resulting dimension and lag.

    Parameters
    ----------
    x : array, shape (n_times - n_times_atom + 1, )
        The array for embedding.
    dim : int
        The dimension of the stride ?

    Returns
    -------
    X.T : array, shape (n_times - 2* (n_times_atom + 1))
        The embedded array
    """
    x = x.copy()
    X = np.lib.stride_tricks.as_strided(x, (len(x) - dim * lag + lag, dim),
                                        (x.strides[0], x.strides[0] * lag))
    return X.T


def solve_unit_norm_dual(lhs, rhs, lambd0, factr=1e7, debug=False,
                         lhs_is_toeplitz=False):
    if np.all(rhs == 0):
        return np.zeros(lhs.shape[0]), 0.

    n_atoms = lambd0.shape[0]
    n_times_atom = lhs.shape[0] // n_atoms

    # precompute SVD
    # U, s, V = linalg.svd(lhs)

    if lhs_is_toeplitz:
        # first column of the toeplitz matrix lhs
        lhs_c = lhs[0, :]

        # lhs will not stay toeplitz if we add different lambd on the diagonal
        assert n_atoms == 1

        def x_star(lambd):
            lambd += 1e-14  # avoid numerical issues
            # lhs_inv = np.dot(V.T / (s + np.repeat(lambd, n_times_atom)), U.T)
            # return np.dot(lhs_inv, rhs)
            lhs_c_copy = lhs_c.copy()
            lhs_c_copy[0] += lambd
            return linalg.solve_toeplitz(lhs_c_copy, rhs)

    else:

        def x_star(lambd):
            lambd += 1e-14  # avoid numerical issues
            # lhs_inv = np.dot(V.T / (s + np.repeat(lambd, n_times_atom)), U.T)
            # return np.dot(lhs_inv, rhs)
            return linalg.solve(lhs + np.diag(np.repeat(lambd, n_times_atom)),
                                rhs)

    def dual(lambd):
        x_hats = x_star(lambd)
        norms = linalg.norm(x_hats.reshape(-1, n_times_atom), axis=1)
        return (x_hats.T.dot(lhs).dot(x_hats) - 2 * rhs.T.dot(x_hats) + np.dot(
            lambd, norms ** 2 - 1.))

    def grad_dual(lambd):
        x_hats = x_star(lambd).reshape(-1, n_times_atom)
        return linalg.norm(x_hats, axis=1) ** 2 - 1.

    def func(lambd):
        return -dual(lambd)

    def grad(lambd):
        return -grad_dual(lambd)

    bounds = [(0., None) for idx in range(0, n_atoms)]
    if debug:
        assert optimize.check_grad(func, grad, lambd0) < 1e-5
    lambd_hats, _, _ = optimize.fmin_l_bfgs_b(func, x0=lambd0, fprime=grad,
                                              bounds=bounds, factr=factr)
    x_hat = x_star(lambd_hats)
    return x_hat, lambd_hats


def solve_unit_norm_primal(lhs, rhs, d_hat0, step_size=0.1, max_iter=1000,
                           factr=1e11, verbose=0):
    """Solve d_hat in the primal using projected gradient descent.

    Stop when (f^k - f^{k+1})/max({|f^k|,|f^{k+1}|,1}) <= factr * eps
    where eps is machine precision.
    """

    def func(d_hat):
        return 0.5 * d_hat.T.dot(lhs).dot(d_hat) - rhs.T.dot(d_hat)

    def grad(d_hat):
        return np.dot(lhs, d_hat) - rhs

    def project(d_hat):
        d_hat /= max(1., linalg.norm(d_hat))
        return d_hat

    eps = np.finfo(np.float64).eps
    d_hat = d_hat0.copy()
    f_last = func(d_hat)
    for ii in range(max_iter):
        d_hat -= step_size * grad(d_hat)
        d_hat = project(d_hat)
        f = func(d_hat)
        if (f_last - f) / max([abs(f), abs(f_last), 1]) <= factr * eps:
            break
        f_last = f
    if verbose > 1:
        print('%d iterations' % (ii + 1))
    return d_hat



# Author: Mathieu Blondel
# License: Simplified BSD

import numpy as np


def delannoy_num(m, n):
    """
    Number of paths from the southwest corner (0, 0) of a rectangular grid to
    the northeast corner (m, n), using only single steps north, northeast, or
    east.

    Named after French army officer and amateur mathematician Henri Delannoy.

    Parameters
    ----------
    m, n : int, int
        Northeast corner coordinates.

    Returns
    -------
    delannoy_num: int
        Delannoy number.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Delannoy_number
    """
    a = np.zeros([m+1, n+1])
    a[0, 0] = 1

    for i in range(1, m+1):
        a[i, 0] = 1

    for j in range(1, n+1):
        a[0, j] = 1

    for i in range(1, m+1):
        for j in range(1, n+1):
            a[i, j] = a[i-1, j] + a[i, j-1] + a[i-1, j-1]

    return a[m, n]


def gen_all_paths(m, n, start=None, M=None):
    """
    Generator that produces all possible paths between (1, 1) and (m, n), using
    only north, northeast, or east steps. Each path is represented as a (m, n)
    numpy array with ones indicating the path.

    Parameters
    ----------
    m, n : int, int
        Northeast corner coordinates.
    """
    if start is None:
        start = [0, 0]
        M = np.zeros((m, n))

    i, j = start
    M[i, j] = 1

    if i == m-1 and j == n-1:
        yield M
    else:
        if i < m - 1:
            # Can use yield_from starting from Python 3.3.
            for mat in gen_all_paths(m, n, (i+1, j), M.copy()):
                yield mat
        if i < m-1 and j < n-1:
            for mat in gen_all_paths(m, n, (i+1, j+1), M.copy()):
                yield mat
        if j < n-1:
            for mat in gen_all_paths(m, n, (i, j+1), M.copy()):
                yield mat



"""
Module profile_this: decorator that profiles a function
"""
import cProfile


def profile_this(fn):
    def profiled_fn(*args, **kwargs):
        filename = fn.__name__ + '.profile'
        prof = cProfile.Profile()
        ret = prof.runcall(fn, *args, **kwargs)
        prof.dump_stats(filename)
        return ret

    return profiled_fn


if __name__ == '__main__':

    def f1():
        a = [x * x for x in range(10000)]
        return a

    def f2():
        a = [x * x for x in range(20000)]
        return a

    def f3():
        a = [x * x for x in range(30000)]
        return a

    @profile_this
    def test():
        f1()
        f2()
        f3()

    test()
    print('Profile is available in test.profile.'
          'Use runsnake or snakeviz to view it')



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

import numpy as np
cimport numpy as cnp
cnp.import_array()

def _locally_greedy_coordinate_selection(z_i, t_start, t_end, tol):
    """Greedily select a coordiante on the segment t_start:t_end for lil matrix

    Parameters
    ----------
    z_i : scipy.sparse.lil_matrix
        The array to select the coordinate from
    t_start, t_end: int
        The segment boundaries for the coordinate selection
    tol : float
        The minimal coordinate update selected

    """
    return _select_argmax_segment(z_i.data, z_i.rows, t_start, t_end, tol)

cdef _select_argmax_segment(object[:] z_i_data,
                            object[:] z_i_rows,
                            int t_start,
                            int t_end,
                            double tol):

    cdef int k, t, i_slice
    cdef int k0 = -1, t0 = -1
    cdef double zk, dz, adz = tol
    for k, (z_k, z_k_row) in enumerate(zip(z_i_data, z_i_rows)):
        for zk, t in zip(z_k, z_k_row):
            if t < t_start:
                continue
            if t >= t_end:
                break
            if abs(zk) > adz:
                k0 = k
                t0 = t
                adz = abs(zk)
                dz = zk

    return k0, t0, dz



def update_dz_opt(z_i, beta, dz_opt, norm_D, reg, t_start, t_end):
    """Update dz_opt for z_i encoded as lil_matrix

    Parameters
    ----------
    z_i : scipy.sparse.lil_matrix
        The array to select the coordinate from
    dz_opt, beta: ndarray, shape (n_atoms, n_times_valid)
        The auxillariay variable to update
    t_start, t_end: int
        The segment boundaries for the coordinate selection

    """
    return _update_dz_opt(z_i.data, z_i.rows, beta, dz_opt, norm_D, reg, t_start,
                          t_end)


cdef _update_dz_opt(object[:] z_i_data,
                    object[:] z_i_rows,
                    cnp.ndarray[double, ndim=2] beta,
                    cnp.ndarray[double, ndim=2] dz_opt,
                    cnp.ndarray[double, ndim=1] norm_D,
                    double reg, int t_start, int t_end):
    cdef double zk, tmp
    cdef int k, t, tk

    for k, (z_k, z_k_row) in enumerate(zip(z_i_data, z_i_rows)):
        norm_Dk = norm_D[k]
        current_t = t_start
        for tk, zk in zip(z_k_row, z_k):
            if tk < t_start:
                continue
            elif tk >= t_end:
                break
            for t in range(current_t, tk):
                tmp = max(-beta[k, t] - reg, 0) / norm_Dk
                dz_opt[k, t] = tmp
            tmp = max(-beta[k, tk] - reg, 0) / norm_Dk
            dz_opt[k, tk] = tmp - zk
            current_t = tk + 1
        for t in range(current_t, t_end):
            tmp = max(-beta[k, t] - reg, 0) / norm_Dk
            dz_opt[k, t] = tmp
    return dz_opt


def subtract_zhat_to_beta(beta, z_hat, norm_Dk):
    """Equivalent to:

    for k, t in zip(*z_hat.nonzero()):
        beta[k, t] -= z_hat[k, t] * norm_Dk[k]
    """
    return _subtract_zhat_to_beta(beta, z_hat.data, z_hat.rows, norm_Dk)


cdef _subtract_zhat_to_beta(cnp.ndarray[double, ndim=2] beta,
                            cnp.ndarray[list, ndim=1] z_i_data,
                            cnp.ndarray[list, ndim=1] z_i_rows,
                            cnp.ndarray[double, ndim=1] norm_Dk):
    cdef double z
    cdef int k, t

    for k, (z_k, z_k_row) in enumerate(zip(z_i_data, z_i_rows)):
        for t, z in zip(z_k_row, z_k):
            beta[k, t] -= z * norm_Dk[k]

    return beta



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Thomas Moreau <thomas.moreau@inria.fr>

import numpy as np

from . import cython_code
from .utils.lil import get_z_shape, is_list_of_lil
from .utils.optim import fista, power_iteration
from .utils.convolution import numpy_convolve_uv
from .utils.compute_constants import compute_ztz, compute_ztX
from .utils.dictionary import tukey_window

from .loss_and_gradient import compute_objective, compute_X_and_objective_multi
from .loss_and_gradient import gradient_uv, gradient_d


def prox_uv(uv, uv_constraint='joint', n_channels=None, return_norm=False):
    if uv_constraint == 'joint':
        norm_uv = np.maximum(1, np.linalg.norm(uv, axis=1))
        uv /= norm_uv[:, None]

    elif uv_constraint == 'separate':
        assert n_channels is not None
        norm_u = np.maximum(1, np.linalg.norm(uv[:, :n_channels], axis=1))
        norm_v = np.maximum(1, np.linalg.norm(uv[:, n_channels:], axis=1))

        uv[:, :n_channels] /= norm_u[:, None]
        uv[:, n_channels:] /= norm_v[:, None]
        norm_uv = norm_u * norm_v
    else:
        raise ValueError('Unknown uv_constraint: %s.' % (uv_constraint, ))

    if return_norm:
        return uv, norm_uv
    else:
        return uv


def prox_d(D, return_norm=False):
    norm_d = np.maximum(1, np.linalg.norm(D, axis=(1, 2)))
    D /= norm_d[:, None, None]

    if return_norm:
        return D, norm_d
    else:
        return D


def update_uv(X, z, uv_hat0, constants=None, b_hat_0=None, debug=False,
              max_iter=300, eps=None, solver_d='alternate', momentum=False,
              uv_constraint='separate', loss='l2', loss_params=dict(),
              verbose=0, window=False):
    """Learn d's in time domain.

    Parameters
    ----------
    X : array, shape (n_trials, n_channels, n_times)
        The data for sparse coding
    z : array, shape (n_trials, n_atoms, n_times - n_times_atom + 1)
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The code for which to learn the atoms
    uv_hat0 : array, shape (n_atoms, n_channels + n_times_atom)
        The initial atoms.
    constants : dict or None
        Dictionary of constants to accelerate the computation of the gradients.
        It should only be given for loss='l2' and should contain ztz and ztX.
    b_hat_0 : array, shape (n_atoms * (n_channels + n_times_atom))
        Init eigen-vector vector used in power_iteration, used in warm start.
    debug : bool
        If True, return the cost at each iteration.
    momentum : bool
        If True, use an accelerated version of the proximal gradient descent.
    uv_constraint : str in {'joint', 'separate'}
        The kind of norm constraint on the atoms:
        If 'joint', the constraint is norm_2([u, v]) <= 1
        If 'separate', the constraint is norm_2(u) <= 1 and norm_2(v) <= 1
    solver_d : str in {'alternate', 'joint'}
        The type of solver to update d:
        If 'alternate', the solver alternates between u then v
        If 'joint', the solver jointly optimize uv with a line search
    loss : str in {'l2' | 'dtw' | 'whitening'}
        The data-fit
    loss_params : dict
        Parameters of the loss
    verbose : int
        Verbosity level.
    window : boolean
        If True, reparametrize the atoms with a temporal Tukey window.

    Returns
    -------
    uv_hat : array, shape (n_atoms, n_channels + n_times_atom)
        The atoms to learn from the data.
    """
    n_trials, n_atoms, n_times_valid = get_z_shape(z)
    _, n_channels, n_times = X.shape
    n_times_atom = uv_hat0.shape[1] - n_channels

    if window:
        tukey_window_ = tukey_window(n_times_atom)[None, :]
        uv_hat0 = uv_hat0.copy()
        uv_hat0[:, n_channels:] /= tukey_window_

    if solver_d == 'alternate':
        msg = "alternate solver should be used with separate constraints"
        assert uv_constraint == 'separate', msg

    if loss == 'l2' and constants is None:
        constants = _get_d_update_constants(X, z)

    def objective(uv):
        if window:
            uv = uv.copy()
            uv[:, n_channels:] *= tukey_window_
        if loss == 'l2':
            return compute_objective(D=uv, constants=constants)
        return compute_X_and_objective_multi(X, z, D_hat=uv, loss=loss,
                                             loss_params=loss_params,
                                             feasible_evaluation=True,
                                             uv_constraint=uv_constraint)

    if solver_d in ['joint', 'fista']:
        # use FISTA on joint [u, v], with an adaptive step size

        def grad(uv):
            if window:
                uv = uv.copy()
                uv[:, n_channels:] *= tukey_window_
            grad = gradient_uv(uv=uv, X=X, z=z, constants=constants, loss=loss,
                               loss_params=loss_params)
            if window:
                grad[:, n_channels:] *= tukey_window_
            return grad

        def prox(uv):
            if window:
                uv[:, n_channels:] *= tukey_window_
            uv = prox_uv(uv, uv_constraint=uv_constraint,
                         n_channels=n_channels)
            if window:
                uv[:, n_channels:] /= tukey_window_
            return uv

        uv_hat, pobj = fista(objective, grad, prox, None, uv_hat0, max_iter,
                             verbose=verbose, momentum=momentum, eps=eps,
                             adaptive_step_size=True, debug=debug,
                             name="Update uv")

    elif solver_d in ['alternate', 'alternate_adaptive']:
        # use FISTA on alternate u and v

        adaptive_step_size = (solver_d == 'alternate_adaptive')

        uv_hat = uv_hat0.copy()
        u_hat, v_hat = uv_hat[:, :n_channels], uv_hat[:, n_channels:]

        def prox_u(u):
            u /= np.maximum(1., np.linalg.norm(u, axis=1))[:, None]
            return u

        def prox_v(v):
            if window:
                v *= tukey_window_
            v /= np.maximum(1., np.linalg.norm(v, axis=1))[:, None]
            if window:
                v /= tukey_window_
            return v

        pobj = []
        for jj in range(1):
            # ---------------- update u

            def obj(u):
                uv = np.c_[u, v_hat]
                return objective(uv)

            def grad_u(u):
                if window:
                    uv = np.c_[u, v_hat * tukey_window_]
                else:
                    uv = np.c_[u, v_hat]
                grad_d = gradient_d(uv, X=X, z=z, constants=constants,
                                    loss=loss, loss_params=loss_params)
                return (grad_d * uv[:, None, n_channels:]).sum(axis=2)

            if adaptive_step_size:
                Lu = 1
            else:
                Lu = compute_lipschitz(uv_hat, constants, 'u', b_hat_0)
            assert Lu > 0

            u_hat, pobj_u = fista(obj, grad_u, prox_u, 0.99 / Lu, u_hat,
                                  max_iter, momentum=momentum, eps=eps,
                                  adaptive_step_size=adaptive_step_size,
                                  verbose=verbose, debug=debug,
                                  name="Update u")
            uv_hat = np.c_[u_hat, v_hat]
            if debug:
                pobj.extend(pobj_u)

            # ---------------- update v
            def obj(v):
                uv = np.c_[u_hat, v]
                return objective(uv)

            def grad_v(v):
                if window:
                    v = v * tukey_window_
                uv = np.c_[u_hat, v]
                grad_d = gradient_d(uv, X=X, z=z, constants=constants,
                                    loss=loss, loss_params=loss_params)
                grad_v = (grad_d * uv[:, :n_channels, None]).sum(axis=1)
                if window:
                    grad_v *= tukey_window_
                return grad_v

            if adaptive_step_size:
                Lv = 1
            else:
                Lv = compute_lipschitz(uv_hat, constants, 'v', b_hat_0)
            assert Lv > 0

            v_hat, pobj_v = fista(obj, grad_v, prox_v, 0.99 / Lv, v_hat,
                                  max_iter, momentum=momentum, eps=eps,
                                  adaptive_step_size=adaptive_step_size,
                                  verbose=verbose, debug=debug,
                                  name="Update v")
            uv_hat = np.c_[u_hat, v_hat]
            if debug:
                pobj.extend(pobj_v)

    else:
        raise ValueError('Unknown solver_d: %s' % (solver_d, ))

    if window:
        uv_hat[:, n_channels:] *= tukey_window_

    if debug:
        return uv_hat, pobj
    return uv_hat


def update_d(X, z, D_hat0, constants=None, b_hat_0=None, debug=False,
             max_iter=300, eps=None, solver_d='fista', momentum=False,
             uv_constraint='joint', loss='l2', loss_params=dict(), verbose=0,
             window=False):
    """Learn d's in time domain.

    Parameters
    ----------
    X : array, shape (n_trials, n_channels, n_times)
        The data for sparse coding
    z : array, shape (n_trials, n_atoms, n_times - n_times_atom + 1)
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The code for which to learn the atoms
    D_hat0 : array, shape (n_atoms, n_channels, n_times_atom)
        The initial atoms.
    constants : dict or None
        Dictionary of constants to accelerate the computation of the gradients.
        It should only be given for loss='l2' and should contain ztz and ztX.
    b_hat_0 : array, shape (n_atoms * (n_channels + n_times_atom))
        Init eigen-vector vector used in power_iteration, used in warm start.
    debug : bool
        If True, return the cost at each iteration.
    momentum : bool
        If True, use an accelerated version of the proximal gradient descent.
    solver_d : str in {'fista'}
        The type of solver to update d:
        If 'fista', the solver optimize D with fista and line search
    loss : str in {'l2' | 'dtw' | 'whitening'}
        The data-fit
    loss_params : dict
        Parameters of the loss
    verbose : int
        Verbosity level.
    window : boolean
        If True, reparametrize the atoms with a temporal Tukey window.

    Returns
    -------
    D_hat : array, shape (n_atoms, n_channels, n_times_atom)
        The atoms to learn from the data.
    """
    n_trials, n_atoms, n_times_valid = get_z_shape(z)
    _, n_channels, n_times = X.shape
    n_atoms, n_channels, n_times_atom = D_hat0.shape

    if window:
        tukey_window_ = tukey_window(n_times_atom)[None, None, :]
        D_hat0 = D_hat0 / tukey_window_

    if loss == 'l2' and constants is None:
        constants = _get_d_update_constants(X, z)

    def objective(D, full=False):
        if window:
            D = D * tukey_window_
        if loss == 'l2':
            return compute_objective(D=D, constants=constants)
        return compute_X_and_objective_multi(X, z, D_hat=D, loss=loss,
                                             loss_params=loss_params)

    if True:  # only solver available here
        # use FISTA on joint [u, v], with an adaptive step size

        def grad(D):
            if window:
                D = D * tukey_window_
            grad = gradient_d(D=D, X=X, z=z, constants=constants, loss=loss,
                              loss_params=loss_params)
            if window:
                grad *= tukey_window_
            return grad

        def prox(D):
            if window:
                D *= tukey_window_
            D = prox_d(D)
            if window:
                D /= tukey_window_
            return D

        D_hat, pobj = fista(objective, grad, prox, None, D_hat0, max_iter,
                            verbose=verbose, momentum=momentum, eps=eps,
                            adaptive_step_size=True, debug=debug,
                            name="Update D")

    else:
        raise ValueError('Unknown solver_d: %s' % (solver_d, ))

    if window:
        D_hat = D_hat * tukey_window_

    if debug:
        return D_hat, pobj
    return D_hat


def _get_d_update_constants(X, z):
    n_trials, n_atoms, n_times_valid = get_z_shape(z)
    n_trials, n_channels, n_times = X.shape
    n_times_atom = n_times - n_times_valid + 1

    if is_list_of_lil(z):
        cython_code._assert_cython()
        ztX = cython_code._fast_compute_ztX(z, X)
        ztz = cython_code._fast_compute_ztz(z, n_times_atom)
    else:
        ztX = compute_ztX(z, X)
        ztz = compute_ztz(z, n_times_atom)

    constants = {}
    constants['ztX'] = ztX
    constants['ztz'] = ztz
    constants['n_channels'] = X.shape[1]
    constants['XtX'] = np.dot(X.ravel(), X.ravel())
    return constants


def compute_lipschitz(uv0, constants, variable, b_hat_0=None):

    n_channels = constants['n_channels']
    u0, v0 = uv0[:, :n_channels], uv0[:, n_channels:]
    n_atoms = uv0.shape[0]
    n_times_atom = uv0.shape[1] - n_channels
    if b_hat_0 is None:
        b_hat_0 = np.random.randn(uv0.size)

    def op_Hu(u):
        u = np.reshape(u, (n_atoms, n_channels))
        uv = np.c_[u, v0]
        H_d = numpy_convolve_uv(constants['ztz'], uv)
        H_u = (H_d * uv[:, None, n_channels:]).sum(axis=2)
        return H_u.ravel()

    def op_Hv(v):
        v = np.reshape(v, (n_atoms, n_times_atom))
        uv = np.c_[u0, v]
        H_d = numpy_convolve_uv(constants['ztz'], uv)
        H_v = (H_d * uv[:, :n_channels, None]).sum(axis=1)
        return H_v.ravel()

    if variable == 'u':
        b_hat_u0 = b_hat_0.reshape(n_atoms, -1)[:, :n_channels].ravel()
        n_points = n_atoms * n_channels
        L = power_iteration(op_Hu, n_points, b_hat_0=b_hat_u0)
    elif variable == 'v':
        b_hat_v0 = b_hat_0.reshape(n_atoms, -1)[:, n_channels:].ravel()
        n_points = n_atoms * n_times_atom
        L = power_iteration(op_Hv, n_points, b_hat_0=b_hat_v0)
    return L



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>

from __future__ import print_function
import time

import numpy as np
from scipy import linalg
from joblib import Parallel

from .init_dict import init_dictionary
from .utils import construct_X, check_random_state
from .utils.dictionary import get_lambda_max
from .update_z import update_z
from .update_d import update_d_block


def objective(X, X_hat, z_hat, reg, sample_weights=None):
    residual = X - X_hat
    if sample_weights is not None:
        residual *= np.sqrt(sample_weights)
    obj = 0.5 * linalg.norm(residual, 'fro') ** 2 + reg * z_hat.sum()
    return obj


def compute_X_and_objective(X, z_hat, d_hat, reg, sample_weights=None,
                            feasible_evaluation=True):
    X_hat = construct_X(z_hat, d_hat)

    if feasible_evaluation:
        z_hat = z_hat.copy()
        d_hat = d_hat.copy()
        # project to unit norm
        d_norm = np.linalg.norm(d_hat, axis=1)
        mask = d_norm >= 1
        d_hat[mask] /= d_norm[mask][:, None]
        # update z in the opposite way
        z_hat[mask] *= d_norm[mask][:, None, None]

    return objective(X, X_hat, z_hat, reg, sample_weights)


def learn_d_z(X, n_atoms, n_times_atom, func_d=update_d_block, reg=0.1,
              lmbd_max='fixed', n_iter=60, random_state=None, n_jobs=1,
              solver_z='l-bfgs', solver_d_kwargs=dict(),
              solver_z_kwargs=dict(), ds_init=None, ds_init_params=dict(),
              sample_weights=None, verbose=10, callback=None,
              stopping_pobj=None):
    """Univariate Convolutional Sparse Coding.

    Parameters
    ----------
    X : array, shape (n_trials, n_times)
        The data on which to perform CSC.
    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    func_d : callable
        The function to update the atoms.
    reg : float
        The regularization parameter
    lmbd_max : 'fixed' | 'scaled' | 'per_atom' | 'shared'
        If not fixed, adapt the regularization rate as a ratio of lambda_max:
          - 'scaled': the regularization parameter is fixed as a ratio of its
            maximal value at init i.e. reg_used = reg * lmbd_max(uv_init)
          - 'shared': the regularization parameter is set at each iteration as
            a ratio of its maximal value for the current dictionary estimate
            i.e. reg_used = reg * lmbd_max(uv_hat)
          - 'per_atom': the regularization parameter is set per atom and at
            each iteration as a ratio of its maximal value for this atom i.e.
            reg_used[k] = reg * lmbd_max(uv_hat[k])
    n_iter : int
        The number of coordinate-descent iterations.
    random_state : int | None
        The random state.
    n_jobs : int
        The number of parallel jobs.
    solver_z : str
        The solver to use for the z update. Options are
        'l-bfgs' (default) | 'ista' | 'fista'
    solver_d_kwargs : dict
        Additional keyword arguments to provide to update_d
    solver_z_kwargs : dict
        Additional keyword arguments to pass to update_z
    ds_init : str or array, shape (n_atoms, n_times_atom)
        The initial atoms or an initialization scheme in {'kmeans' | 'ssa' |
        'chunks' | 'random'}.
    ds_init_params : dict
        Dictionnary of parameters for the kmeans init method.
    sample_weights : array, shape (n_trials, n_times)
        The weights in the alphaCSC problem. Should be None
        when using vanilla CSC.
    verbose : int
        The verbosity level.
    callback : func
        A callback function called at the end of each loop of the
        coordinate descent.

    Returns
    -------
    pobj : list
        The objective function value at each step of the coordinate descent.
    times : list
        The cumulative time for each iteration of the coordinate descent.
    d_hat : array, shape (n_atoms, n_times)
        The estimated atoms.
    z_hat : array, shape (n_atoms, n_trials, n_times - n_times_atom + 1)
        The sparse activation matrix.
    reg : float
        Regularization parameter used.
    """
    n_trials, n_times = X.shape

    rng = check_random_state(random_state)

    # reuse multivariate atoms init function
    if isinstance(ds_init, np.ndarray):
        ds_init = ds_init[:, None, :]
    d_hat = init_dictionary(X[:, None, :], n_atoms, n_times_atom,
                            D_init=ds_init, rank1=False,
                            D_init_params=ds_init_params, random_state=rng)
    d_hat = d_hat[:, 0, :]

    # strategy for rescaling the regularization parameter
    reg0 = reg
    lambda_max = get_lambda_max(X, d_hat, sample_weights).max()
    if lmbd_max == "scaled":
        reg = reg0 * lambda_max

    pobj = list()
    times = list()

    if 'ista' in solver_z:
        b_hat_0 = rng.randn(n_atoms * (n_times - n_times_atom + 1))
    else:
        b_hat_0 = None

    lambd0 = None
    z_hat = np.zeros((n_atoms, n_trials, n_times - n_times_atom + 1))

    pobj.append(compute_X_and_objective(X, z_hat, d_hat, reg, sample_weights))
    times.append(0.)
    with Parallel(n_jobs=n_jobs) as parallel:
        for ii in range(n_iter):  # outer loop of coordinate descent
            if verbose == 1:
                msg = '.' if (ii % 50 != 0) else 'V_%d/%d ' % (ii, n_iter)
                print(msg, end='', flush=True)
            if verbose > 1:
                print('Coordinate descent loop %d / %d [n_jobs=%d]' %
                      (ii, n_iter, n_jobs))

            if lmbd_max not in ['fixed', 'scaled']:
                lambda_max = get_lambda_max(X, d_hat, sample_weights)
                reg = reg0 * lambda_max
                if lmbd_max == 'shared':
                    reg = reg.max()

            start = time.time()
            z_hat = update_z(X, d_hat, reg, z0=z_hat, parallel=parallel,
                             solver=solver_z, b_hat_0=b_hat_0,
                             solver_kwargs=solver_z_kwargs,
                             sample_weights=sample_weights)
            times.append(time.time() - start)

            # monitor cost function
            pobj.append(
                compute_X_and_objective(X, z_hat, d_hat, reg, sample_weights))
            if verbose > 1:
                print('[seed %s] Objective (z_hat) : %0.8f' % (random_state,
                                                               pobj[-1]))

            if len(z_hat.nonzero()[0]) == 0:
                import warnings
                warnings.warn("Regularization parameter `reg` is too large "
                              "and all the activations are zero. No atoms has"
                              " been learned.", UserWarning)
                break

            start = time.time()
            d_hat, lambd0 = func_d(X, z_hat, n_times_atom, lambd0=lambd0,
                                   ds_init=d_hat, verbose=verbose,
                                   solver_kwargs=solver_d_kwargs,
                                   sample_weights=sample_weights)
            times.append(time.time() - start)

            # monitor cost function
            pobj.append(
                compute_X_and_objective(X, z_hat, d_hat, reg, sample_weights))
            if verbose > 1:
                print('[seed %s] Objective (d) %0.8f' % (random_state,
                                                         pobj[-1]))

            if callable(callback):
                callback(X, d_hat, z_hat, reg)

            if stopping_pobj is not None and pobj[-1] < stopping_pobj:
                break

        if verbose == 1:
            print('')

    return pobj, times, d_hat, z_hat, reg



"""
===========================
Alpha CSC on simulated data
===========================

This example demonstrates alphaCSC [1]_ on simulated data.

.. [1] Jas, M., Dupré La Tour, T., Şimşekli, U., & Gramfort, A. (2017).
    `Learning the Morphology of Brain Signals Using Alpha-Stable Convolutional
    Sparse Coding
    <https://papers.nips.cc/paper/6710-learning-the-morphology-of-brain-signals-using-alpha-stable-convolutional-sparse-coding.pdf>`_.
    Advances in Neural Information Processing Systems (NIPS), pages 1099--1108.
"""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first define the parameters of our model.

n_times_atom = 64  # L
n_times = 512  # T
n_atoms = 2  # K
n_trials = 100  # N

alpha = 1.2
reg = 0.1

###############################################################################
# Next, we define the parameters for alpha CSC

n_iter_global = 10
n_iter_optim = 50
n_iter_mcmc = 100
n_burnin_mcmc = 50

###############################################################################
# Here, we simulate the data

from alphacsc.simulate import simulate_data

random_state_simulate = 1
X, ds_true, z_true = simulate_data(n_trials, n_times, n_times_atom,
                                   n_atoms, random_state_simulate)

###############################################################################
# Add some noise and corrupt some trials even with impulsive noise

from scipy.stats import levy_stable
from alphacsc import check_random_state

fraction_corrupted = 0.02
n_corrupted_trials = int(fraction_corrupted * n_trials)

# Add stationary noise:
rng = check_random_state(random_state_simulate)
X += 0.01 * rng.randn(*X.shape)

noise_level = 0.005
# add impulsive noise
idx_corrupted = rng.randint(0, n_trials,
                            size=n_corrupted_trials)
X[idx_corrupted] += levy_stable.rvs(alpha, 0, loc=0, scale=noise_level,
                                    size=(n_corrupted_trials, n_times),
                                    random_state=random_state_simulate)

###############################################################################
# and then alpha CSC on the same data

from alphacsc import learn_d_z_weighted

d_hat, z_hat, Tau = learn_d_z_weighted(
    X, n_atoms, n_times_atom, reg=reg, alpha=alpha,
    solver_d_kwargs=dict(factr=100), n_iter_global=n_iter_global,
    n_iter_optim=n_iter_optim, init_tau=True,
    n_iter_mcmc=n_iter_mcmc, n_burnin_mcmc=n_burnin_mcmc,
    random_state=60, n_jobs=1, verbose=1)

###############################################################################
# Finally, let's compare the results. Now, it works even in the presence
# of impulsive noise.

import matplotlib.pyplot as plt

plt.figure()
plt.plot(d_hat.T, 'b', label=r'$\alpha$CSC')
plt.plot(ds_true.T, 'k--', label='True atoms')
handles, labels = plt.gca().get_legend_handles_labels()
plt.legend(handles[::2], labels[::2], loc='best')
plt.show()

###############################################################################
# We can even visualize the weights to see what time points were
# downweighted by the algorithm

fig, axes = plt.subplots(2, 1, sharex=True)
axes[0].set_xlim([-20, n_times])
axes[0].set_ylim([0, 2])
axes[1].set_ylim([-0.5, 0.3])

for t in [2, 250]:
    axes[0].axvline(t, linestyle='-.')
    axes[1].axvline(t, linestyle='-.')

axes[0].plot(1 / Tau[idx_corrupted[0], :])
axes[0].set_ylabel('1 / weights')
axes[1].plot(X[idx_corrupted[0], :])
axes[1].set_ylabel('Corrupted trial')



# -*- coding: utf-8 -*-
"""K-medoids clustering

Taken from https://github.com/scikit-learn/scikit-learn/pull/7694

"""

# Authors: Timo Erkkila <timo.erkkila@gmail.com>
#          Antti Lehmussola <antti.lehmussola@gmail.com>
#          Kornel Kielczewski <kornel.k@plusnet.pl>
# License: BSD 3 clause

import warnings

import numpy as np

from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.utils import check_array, check_random_state
from sklearn.utils.validation import check_is_fitted


class KMedoids(BaseEstimator, ClusterMixin, TransformerMixin):
    """k-medoids clustering.

    Read more in the :ref:`User Guide <k_medoids>`.

    Parameters
    ----------
    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of medoids to
        generate.

    distance_metric : string, optional, default: 'euclidean'
        What distance metric to use.

    init : {'random', 'heuristic'}, optional, default: 'heuristic'
        Specify medoid initialization method. Random selects n_clusters
        elements from the dataset, while heuristic picks the n_clusters points
        with the smallest sum distance to every other point.

    max_iter : int, optional, default : 300
        Specify the maximum number of iterations when fitting.

    random_state : int, optional, default: None
        Specify random state for the random number generator.

    Attributes
    ----------
    cluster_centers_ : array, [n_clusters, n_features]
        Cluster centers, i.e. medoids (elements from the original dataset)

    labels_ :
        Labels of each point

    inertia_ : float
        Sum of distances of samples to their closest cluster center.

    Examples
    --------

    >>> from sklearn.cluster import KMedoids
    >>> import numpy as np

    >>> X = np.asarray([[1, 2], [1, 4], [1, 0],
    ...                 [4, 2], [4, 4], [4, 0]])
    >>> kmedoids = KMedoids(n_clusters=2, random_state=0).fit(X)
    >>> kmedoids.labels_
    array([0, 0, 0, 1, 1, 1])
    >>> kmedoids.predict([[0,0], [4,4]])
    array([0, 1])
    >>> kmedoids.cluster_centers_
    array([[1, 2],
           [4, 2]])
    >>> kmedoids.inertia_
    8.0

    >>> kmedoids = KMedoids(n_clusters=2, random_state=0,
    ...                     distance_metric='manhattan').fit(X)
    >>> kmedoids.labels_
    array([0, 0, 0, 1, 1, 1])
    >>> kmedoids.predict([[0,0], [4,4]])
    array([0, 1])
    >>> kmedoids.cluster_centers_
    array([[1, 2],
           [4, 2]])
    >>> kmedoids.inertia_
    8.0

    References
    ----------
    Kaufman, L. and Rousseeuw, P.J., Statistical Data Analysis Based on
    the L1-Norm and Related Methods, edited by Y. Dodge, North-Holland,
    405-416. 1987
    """

    def __init__(self, n_clusters=8, distance_metric='euclidean',
                 init='heuristic', max_iter=300, random_state=None):
        self.n_clusters = n_clusters
        self.distance_metric = distance_metric
        self.init = init
        self.max_iter = max_iter
        self.random_state = random_state

    def _check_init_args(self):
        """Validates the input arguments. """

        # Check n_clusters
        if (self.n_clusters is None or self.n_clusters <= 0 or
                not isinstance(self.n_clusters, (int, np.integer))):
            raise ValueError("n_clusters should be a nonnegative integer. "
                             "%s was given" % self.n_clusters)

        # Check init
        init_methods = ['random', 'heuristic']
        if isinstance(self.init, str) and self.init not in init_methods:
            raise ValueError("init needs to be one of " + "the following: " +
                             "%s" % init_methods)
        if isinstance(self.init, np.ndarray):
            assert self.init.size == self.n_clusters

        # Check random state
        self.random_state_ = check_random_state(self.random_state)

    def fit(self, X, y=None):
        """Fit K-Medoids to the provided data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=(n_samples, n_features).
            Dataset to cluster.

        Returns
        -------
        self
        """

        self._check_init_args()
        X = check_array(X, accept_sparse=['csr', 'csc'])
        if self.n_clusters > X.shape[0]:
            raise ValueError("The number of medoids %d must be less "
                             "than the number of samples %d." %
                             (self.n_clusters, X.shape[0]))

        # if not isinstance(self.init, np.ndarray):
        if callable(self.distance_metric):
            distances = self.distance_metric(X)
        else:
            distances = pairwise_distances(X, metric=self.distance_metric)
        # else:
        #     distances = None

        medoid_idxs = self._get_initial_medoid_indices(distances,
                                                       self.n_clusters)
        if distances is not None:
            this = distances[medoid_idxs, :]
        else:
            this = self.distance_metric(X, np.atleast_2d(X[medoid_idxs])).T
        labels = np.argmin(this, axis=0)

        # Old medoids will be stored here for reference
        old_medoid_idxs = np.zeros((self.n_clusters, ))

        # Continue the algorithm as long as
        # the medoids keep changing and the maximum number
        # of iterations is not exceeded
        self.n_iter_ = 0
        while (not np.all(old_medoid_idxs == medoid_idxs) and
               self.n_iter_ < self.max_iter):
            self.n_iter_ += 1
            old_medoid_idxs = np.copy(medoid_idxs)
            if distances is not None:
                this = distances[medoid_idxs, :]
            else:
                this = self.distance_metric(X, np.atleast_2d(X[medoid_idxs])).T
            labels = np.argmin(this, axis=0)

            # Update medoids with the new cluster indices
            self._update_medoid_idxs_in_place(X, distances, labels,
                                              medoid_idxs)

        # Expose labels_ which are the assignments of
        # the training data to clusters
        self.labels_ = labels
        self.medoid_idxs_ = medoid_idxs
        self.cluster_centers_ = X[medoid_idxs]
        self.inertia_ = self._compute_inertia(X)

        # Return self to enable method chaining
        return self

    def _update_medoid_idxs_in_place(self, X, distances, cluster_idxs,
                                     medoid_idxs):
        """In-place update of the medoid indices"""

        # Update the medoids for each cluster
        for k in range(self.n_clusters):

            if sum(cluster_idxs == k) == 0:
                warnings.warn("Cluster %d is empty!" % k)
                continue

            # Extract the distance matrix between the data points
            # inside the cluster k
            cluster_k_idxs = np.where(cluster_idxs == k)[0]
            if distances is not None:
                in_cluster_distances = distances[np.ix_(
                    cluster_k_idxs, cluster_k_idxs)]
            else:
                in_cluster_distances = self.distance_metric(
                    np.atleast_2d(X[cluster_k_idxs]))

            # Calculate all costs from each point to all others in the cluster
            in_cluster_all_costs = np.sum(in_cluster_distances, axis=1)

            min_cost_idx = np.argmin(in_cluster_all_costs)
            min_cost = in_cluster_all_costs[min_cost_idx]

            if distances is not None:
                curr_cost = np.sum(distances[medoid_idxs[k], cluster_k_idxs])
            else:
                curr_cost = np.sum(
                    self.distance_metric(
                        np.atleast_2d(X[cluster_k_idxs]),
                        np.atleast_2d(X[medoid_idxs[k]])).T)

            # If the minimum cost is smaller than that
            # exhibited by the currently used medoid,
            # we switch to using the new medoid in cluster k
            if min_cost < curr_cost:
                # Find data points that belong to cluster k,
                # and assign the newly found medoid as the medoid
                # for cluster k
                medoid_idxs[k] = cluster_k_idxs[min_cost_idx]

    def transform(self, X):
        """Transforms X to cluster-distance space.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=(n_samples, n_features)
            Data to transform.

        Returns
        -------
        X_new : {array-like, sparse matrix}, shape=(n_samples, n_clusters)
            X transformed in the new space of distances to cluster centers.
        """
        X = check_array(X, accept_sparse=['csr', 'csc'])
        check_is_fitted(self, "cluster_centers_")

        if callable(self.distance_metric):
            return self.distance_metric(X, Y=self.cluster_centers_)
        else:
            return pairwise_distances(X, Y=self.cluster_centers_,
                                      metric=self.distance_metric)

    def predict(self, X):
        """Predict the closest cluster for each sample in X

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self, "cluster_centers_")
        X = check_array(X, accept_sparse=['csr', 'csc'])

        if callable(self.distance_metric):
            distances = self.distance_metric(X, Y=self.cluster_centers_)
        else:
            distances = pairwise_distances(X, Y=self.cluster_centers_,
                                           metric=self.distance_metric)

        # Assign data points to clusters based on which cluster assignment
        # yields the smallest distance
        labels = np.argmin(distances, axis=1)

        return labels

    def _compute_inertia(self, X):
        """Compute inertia of new samples. Inertia is defined as the sum of the
        sample distances to closest cluster centers.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=(n_samples, n_features)
            Samples to compute inertia for.

        Returns
        -------
        Sum of sample distances to closest cluster centers.
        """
        distances = self.transform(X)

        # Define inertia as the sum of the sample-distances
        # to closest cluster centers
        inertia = np.sum(np.min(distances, axis=1))

        return inertia

    def _get_initial_medoid_indices(self, distances, n_clusters):
        """Select initial medoids randomly or heuristically"""

        if isinstance(self.init, str) and self.init == 'random':
            # Pick random k medoids as the initial ones.
            medoids = self.random_state_.permutation(
                distances.shape[0])[:n_clusters]
        elif isinstance(self.init, str) and self.init == 'heuristic':
            # Pick K first data points that have the smallest sum distance
            # to every other point. These are the initial medoids.
            medoids = list(np.argsort(np.sum(distances, axis=1))[:n_clusters])
        elif isinstance(self.init, np.ndarray):
            medoids = self.init
        else:
            raise ValueError(
                "Initialization not implemented for method: '%s'" % self.init)

        return medoids



from os.path import join
from copy import deepcopy

import mne
import numpy as np
from joblib import Memory
from scipy.signal import tukey

mem = Memory(location='.', verbose=0)


@mem.cache(ignore=['n_jobs'])
def load_data(dataset="somato", n_splits=10, sfreq=None, epoch=None,
              filter_params=[2., None], return_array=True, n_jobs=1):
    """Load and prepare the somato dataset for multiCSC

    Parameters
    ----------
    dataset : str in {'somato', 'sample'}
        Dataset to load.
    n_splits : int
        Split the signal in n_split signals of same length before returning it.
        If epoch is provided, the signal is instead splitted according to the
        epochs and this option is not followed.
    sfreq : float
        Sampling frequency of the signal. The data are resampled to match it.
    epoch : tuple or None
        If set to a tuple, extract epochs from the raw data, using
        t_min=epoch[0] and t_max=epoch[1]. Else, use the raw signal, divided
        in n_splits chunks.
    filter_params : tuple of length 2
        Boundaries of filtering, e.g. (2, None), (30, 40), (None, 40).
    return_array : boolean
        If True, return an NumPy array, instead of mne objects.
    n_jobs : int
        Number of jobs that can be used for preparing (filtering) the data.

    Returns
    -------
    X : array, shape (n_splits, n_channels, n_times)
        The loaded dataset.
    info : dict
        MNE dictionary of information about recording settings.
    """

    pick_types_epoch = dict(meg='grad', eeg=False, eog=True, stim=False)
    pick_types_final = dict(meg='grad', eeg=False, eog=False, stim=False)

    if dataset == 'somato':
        data_path = mne.datasets.somato.data_path()
        subjects_dir = join(data_path, "subjects")
        data_dir = join(data_path, 'MEG', 'somato')
        file_name = join(data_dir, 'sef_raw_sss.fif')
        raw = mne.io.read_raw_fif(file_name, preload=True)
        raw.notch_filter(np.arange(50, 101, 50), n_jobs=n_jobs)
        event_id = 1

        # Dipole fit information
        cov = None  # see below
        file_trans = join(data_dir, "sef_raw_sss-trans.fif")
        file_bem = join(subjects_dir, 'somato', 'bem',
                        'somato-5120-bem-sol.fif')

    elif dataset == 'sample':
        data_path = mne.datasets.sample.data_path()
        subjects_dir = join(data_path, "subjects")
        data_dir = join(data_path, 'MEG', 'sample')
        file_name = join(data_dir, 'sample_audvis_raw.fif')
        raw = mne.io.read_raw_fif(file_name, preload=True)
        raw.notch_filter(np.arange(60, 181, 60), n_jobs=n_jobs)
        event_id = [1, 2, 3, 4]

        # Dipole fit information
        cov = join(data_dir, 'sample_audvis-cov.fif')
        file_trans = join(data_dir, 'sample_audvis_raw-trans.fif')
        file_bem = join(subjects_dir, 'sample', 'bem',
                        'sample-5120-bem-sol.fif')

    else:
        raise ValueError('Unknown parameter dataset=%s.' % (dataset, ))
    raw.filter(*filter_params, n_jobs=n_jobs)

    baseline = (None, 0)
    events = mne.find_events(raw, stim_channel='STI 014')
    events = mne.pick_events(events, include=event_id)

    # compute the covariance matrix for somato
    if dataset == "somato":
        picks_cov = mne.pick_types(raw.info, **pick_types_epoch)
        epochs_cov = mne.Epochs(raw, events, event_id, tmin=-4, tmax=0,
                                picks=picks_cov, baseline=baseline,
                                reject=dict(grad=4000e-13, eog=350e-6),
                                preload=True)
        epochs_cov.pick_types(**pick_types_final)
        cov = mne.compute_covariance(epochs_cov)

    if epoch:
        t_min, t_max = epoch

        picks = mne.pick_types(raw.info, **pick_types_epoch)
        epochs = mne.Epochs(raw, events, event_id, t_min, t_max, picks=picks,
                            baseline=baseline, reject=dict(
                                grad=4000e-13, eog=350e-6), preload=True)
        epochs.pick_types(**pick_types_final)
        info = epochs.info
        if sfreq is not None:
            epochs = epochs.resample(sfreq, npad='auto', n_jobs=n_jobs)

        if return_array:
            X = epochs.get_data()

    else:
        events[:, 0] -= raw.first_samp
        raw.pick_types(**pick_types_final)
        info = raw.info

        if sfreq is not None:
            raw, events = raw.resample(sfreq, events=events, npad='auto',
                                       n_jobs=n_jobs)

        if return_array:
            X = raw.get_data()
            n_channels, n_times = X.shape
            n_times = n_times // n_splits
            X = X[:, :n_splits * n_times]
            X = X.reshape(n_channels, n_splits, n_times).swapaxes(0, 1)

    # Deep copy before modifying info to avoid issues when saving EvokedArray
    info = deepcopy(info)
    info['events'] = events
    info['event_id'] = event_id
    info['subject'] = dataset
    info['subjects_dir'] = subjects_dir

    info['cov'] = cov
    info['file_bem'] = file_bem
    info['file_trans'] = file_trans

    if return_array:
        n_splits, n_channels, n_times = X.shape
        X *= tukey(n_times, alpha=0.1)[None, None, :]
        X /= np.std(X)
        return X, info
    elif epoch:
        return epoch, info
    else:
        return raw, info



"""
==============================
CSC to learn LFP spiking atoms
==============================

Here, we show how CSC can be used to learn spiking
atoms from Local Field Potential (LFP) data [1].

[1] Hitziger, Sebastian, et al.
    Adaptive Waveform Learning: A Framework for Modeling Variability in
    Neurophysiological Signals. IEEE Transactions on Signal Processing (2017).
"""

###############################################################################
# First, let us fetch the data (~14 MB)
import os
from mne.utils import _fetch_file

url = ('https://github.com/hitziger/AWL/raw/master/Experiments/data/'
       'LFP_data_contiguous_1250_Hz.mat')
fname = './LFP_data_contiguous_1250_Hz.mat'
if not os.path.exists(fname):
    _fetch_file(url, fname)

###############################################################################
# It is a mat file, so we use scipy to load it
from scipy import io

data = io.loadmat(fname)
X, sfreq = data['X'].T, float(data['sfreq'])

###############################################################################
# And now let us look at the data
import numpy as np
import matplotlib.pyplot as plt

start, stop = 11000, 15000
times = np.arange(start, stop) / sfreq
plt.plot(times, X[0, start:stop], color='b')
plt.xlabel('Time (s)')
plt.ylabel(r'$\mu$ V')
plt.xlim([9., 12.])

###############################################################################
# and filter it using a convenient function from MNE. This will remove low
# frequency drifts, but we keep the high frequencies
from mne.filter import filter_data
X = filter_data(
    X.astype(np.float64), sfreq, l_freq=1, h_freq=None, fir_design='firwin')

###############################################################################
# Now, we define the parameters of our model.

reg = 6.0
n_times = 2500
n_times_atom = 350
n_trials = 100
n_atoms = 3
n_iter = 60
###############################################################################
# Let's stick to one random state for now, but if you want to learn how to
# select the random state, consult :ref:`this example
# <sphx_glr_auto_examples_plot_simulate_randomstate.py>`.
random_state = 10

###############################################################################
# Now, we epoch the trials

overlap = 0
starts = np.arange(0, X.shape[1] - n_times, n_times - overlap)
stops = np.arange(n_times, X.shape[1], n_times - overlap)

X_new = []
for idx, (start, stop) in enumerate(zip(starts, stops)):
    if idx >= n_trials:
        break
    X_new.append(X[0, start:stop])
X_new = np.vstack(X_new)
del X

###############################################################################
# We remove the mean and scale to unit variance.
X_new -= np.mean(X_new)
X_new /= np.std(X_new)

###############################################################################
# The convolutions can result in edge artifacts at the edges of the trials.
# Therefore, we discount the contributions from the edges by windowing the
# trials.
from numpy import hamming
X_new *= hamming(n_times)[None, :]

###############################################################################
# Of course, in a data-limited setting we want to use as much of the data as
# possible. If this is the case, you can set `overlap` to non-zero (for example
# half the epoch length).
#
# Now, we run regular CSC since the trials are not too noisy
from alphacsc import learn_d_z
pobj, times, d_hat, z_hat, reg = learn_d_z(X_new, n_atoms, n_times_atom,
                                           reg=reg, n_iter=n_iter,
                                           random_state=random_state, n_jobs=1)

###############################################################################
# Let's look at the atoms now.
plt.figure()
plt.plot(d_hat.T)
plt.show()



from .update_d import update_d_block
from .learn_d_z import learn_d_z, objective
from .learn_d_z_mcem import learn_d_z_weighted
from .learn_d_z_multi import learn_d_z_multi
from .utils import construct_X, check_random_state

from .online_dictionary_learning import OnlineCDL
from .convolutional_dictionary_learning import BatchCDL, GreedyCDL

__all__ = [
    "BatchCDL",
    "GreedyCDL",
    "OnlineCDL",
    "construct_X",
    "check_random_state",
    "learn_d_z",
    "learn_d_z_multi",
    "learn_d_z_weighted",
    "objective",
    "update_d_block",
]



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Thomas Moreau <thomas.moreau@inria.fr>

from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError

from .update_z_multi import update_z_multi
from .utils.dictionary import get_D, get_uv
from .learn_d_z_multi import learn_d_z_multi
from .loss_and_gradient import construct_X_multi


DOC_FMT = """{desc}

    Parameters
    ----------

    Problem Specs

    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    loss : {{ 'l2' | 'dtw' | 'whitening' }}
        Loss for the data-fit term. Either the norm l2 or the soft-DTW.
    loss_params : dict
        Parameters of the loss.
    rank1 : boolean
        If set to True, learn rank 1 dictionary atoms.
    window : boolean
        If set to True, re-parametrizes the atoms with a temporal Tukey window.
    uv_constraint : {{'joint' | 'separate'}}
        The kind of norm constraint on the atoms:

        - :code:`'joint'`: the constraint is ||[u, v]||_2 <= 1
        - :code:`'separate'`: the constraint is ||u||_2 <= 1 and ||v||_2 <= 1
    sort_atoms : boolean
        If True, the atoms are sorted by explained variances.


    Global algorithm
    {algorithm}
    n_iter : int
        The number of alternate steps to perform.
    eps : float
        Stopping criterion. If the cost descent after a uv and a z update is
        smaller than eps, return.
    reg : float
        The regularization parameter.
    lmbd_max : 'fixed' | 'scaled' | 'per_atom' | 'shared'
        If not fixed, adapt the regularization rate as a ratio of lambda_max:

        - :code:`'scaled'`: the regularization parameter is fixed as a ratio of
          its maximal value at init *i.e.*
          :math:`lambda` = reg * lmbd_max(uv_init).
        - :code:`'shared'`: the regularization parameter is set at each
          iteration as a ratio of its maximal value for the current dictionary
          estimate *i.e.* :math:`lambda` = reg * lmbd_max(uv_hat).
        - :code:`'per_atom'`: the regularization parameter is set per atom and
          at each iteration as a ratio of its maximal value for this atom
          *i.e.* :math:`lambda[k]` = reg * lmbd_max(uv_hat[k]).


    Z-step parameters

    solver_z : str
        The solver to use for the z update. Options are
        {{'l_bfgs' (default) | 'lgcd'}}.
    solver_z_kwargs : dict
        Additional keyword arguments to pass to update_z_multi.
    use_sparse_z : boolean
        Use sparse lil_matrices to store the activations.
    unbiased_z_hat : boolean
        If set to True, the value of the non-zero coefficients in the returned
        z_hat are recomputed with reg=0 on the frozen support.


    D-step parameters

    solver_d : str
        The solver to use for the d update. Options are
        'alternate' | 'alternate_adaptive' (default) | 'joint'
    solver_d_kwargs : dict
        Additional keyword arguments to provide to update_d
    D_init : str or array
        The initial atoms with shape (n_atoms, n_channels + n_times_atoms) or
        (n_atoms, n_channels, n_times_atom) or an initialization scheme str in
        {{'kmeans' | 'ssa' | 'chunks' | 'random'}}.
    D_init_params : dict
        Dictionnary of parameters for the kmeans init method.


    Technical parameters

    n_jobs : int
        The number of parallel jobs.
    verbose : int
        The verbosity level.
    callback : func
        A callback function called at the end of each loop of the
        coordinate descent.
    random_state : int | None
        State to seed the random number generator.
    raise_on_increase : boolean
        Raise an error if the objective function increase.

    """

DEFAULT = dict(
    desc="Base class for convolutional dictionary learning algorithms",
    algorithm="""

    algorithm : {'batch' | 'greedy' | 'online'}
        Dictionary learning algorithm.
    algorithm_params : dict
        parameter of the global algorithm."""
)


class ConvolutionalDictionaryLearning(TransformerMixin):
    __doc__ = DOC_FMT.format(**DEFAULT)

    def __init__(self, n_atoms, n_times_atom, n_iter=60, n_jobs=1,
                 loss='l2', loss_params=dict(gamma=.1, sakoe_chiba_band=10,
                                             ordar=10),
                 rank1=True, window=False, uv_constraint='separate',
                 solver_z='l_bfgs', solver_z_kwargs={},
                 solver_d='alternate_adaptive', solver_d_kwargs={},
                 reg=0.1, lmbd_max='fixed', eps=1e-10,
                 D_init=None, D_init_params={},
                 algorithm='batch', algorithm_params={},
                 alpha=.8, batch_size=1, batch_selection='random',
                 use_sparse_z=False, unbiased_z_hat=False,
                 verbose=10, callback=None, random_state=None, name="_CDL",
                 raise_on_increase=True, sort_atoms=False):

        # Problem Specs
        self.n_atoms = n_atoms
        self.n_times_atom = n_times_atom
        self.reg = reg
        self.loss = loss
        self.loss_params = loss_params
        self.rank1 = rank1
        self.window = window
        self.uv_constraint = uv_constraint
        self.sort_atoms = sort_atoms

        # Global algorithm
        self.n_iter = n_iter
        self.eps = eps
        self.algorithm = algorithm
        self.algorithm_params = algorithm_params
        self.lmbd_max = lmbd_max

        # Z-step parameters
        self.solver_z = solver_z
        self.solver_z_kwargs = solver_z_kwargs
        self.use_sparse_z = use_sparse_z
        self.unbiased_z_hat = unbiased_z_hat

        # D-step parameters
        self.solver_d = solver_d
        self.solver_d_kwargs = solver_d_kwargs
        self.D_init = D_init
        self.D_init_params = D_init_params

        # Technical parameters
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.callback = callback
        self.random_state = random_state
        self.raise_on_increase = raise_on_increase
        self.name = name

        # Init property
        self._D_hat = None

    def fit(self, X, y=None):
        """Learn a convolutional dictionary from the set of signals X.
        """
        res = learn_d_z_multi(
            X, self.n_atoms, self.n_times_atom,
            reg=self.reg, lmbd_max=self.lmbd_max,
            loss=self.loss, loss_params=self.loss_params,
            rank1=self.rank1, window=self.window,
            uv_constraint=self.uv_constraint,
            algorithm=self.algorithm, algorithm_params=self.algorithm_params,
            n_iter=self.n_iter, eps=self.eps,
            solver_z=self.solver_z, solver_z_kwargs=self.solver_z_kwargs,
            solver_d=self.solver_d, solver_d_kwargs=self.solver_d_kwargs,
            D_init=self.D_init, D_init_params=self.D_init_params,
            use_sparse_z=self.use_sparse_z, unbiased_z_hat=False,
            verbose=self.verbose, callback=self.callback,
            random_state=self.random_state, n_jobs=self.n_jobs,
            name=self.name, raise_on_increase=self.raise_on_increase,
            sort_atoms=self.sort_atoms)

        self._pobj, self._times, self._D_hat, self._z_hat, self.reg_ = res
        self.n_channels_ = X.shape[1]
        return self

    def fit_transform(self, X, y=None):
        """Learn a convolutional dictionary and returns sparse codes.
        """
        self.fit(X)

        z_hat = self._z_hat

        if self.unbiased_z_hat:
            if self.verbose > 0:
                print("Refitting the activation to avoid amplitude bias...")
            z_hat, _, _ = update_z_multi(
                X, self._D_hat, z0=z_hat, n_jobs=self.n_jobs,
                reg=0, freeze_support=True,
                solver=self.solver_z, solver_kwargs=self.solver_z_kwargs,
                loss=self.loss, loss_params=self.loss_params)
            if self.verbose > 0:
                print("done")

        return z_hat

    def transform(self, X):
        """Returns sparse codes associated to the signals X for the dictionary.
        """
        self._check_fitted()
        z_hat, _, _ = update_z_multi(
            X, self._D_hat, reg=self.reg_, n_jobs=self.n_jobs,
            solver=self.solver_z, solver_kwargs=self.solver_z_kwargs,
            loss=self.loss, loss_params=self.loss_params)

        if self.unbiased_z_hat:
            if self.verbose > 0:
                print("Refitting the activation to avoid amplitude bias...")
            z_hat, _, _ = update_z_multi(
                X, self._D_hat, z0=z_hat, n_jobs=self.n_jobs,
                reg=0, freeze_support=True,
                solver=self.solver_z, solver_kwargs=self.solver_z_kwargs,
                loss=self.loss, loss_params=self.loss_params)
            if self.verbose > 0:
                print("done")
        return z_hat

    def transform_inverse(self, z_hat):
        """Reconstruct the signals from the given sparse codes.
        """
        return construct_X_multi(z_hat, self._D_hat, self.n_channels_)

    def _check_fitted(self):
        if self._D_hat is None:
            raise NotFittedError("Fit must be called before accessing the "
                                 "dictionary")

    @property
    def D_hat_(self):
        """array: dictionary in full rank mode.

        shape (n_atoms, n_channels, n_times_atom)
        """
        self._check_fitted()
        if self._D_hat.ndim == 3:
            return self._D_hat

        return get_D(self._D_hat, self.n_channels_)

    @property
    def uv_hat_(self):
        """array: dictionary in rank 1 mode. If `rank1 = False`, this is an
        approximation of the dictionary obtained through svd.

        shape (n_atoms, n_channels + n_times_atom)
        """
        self._check_fitted()
        if self._D_hat.ndim == 3:
            return get_uv(self._D_hat)

        return self._D_hat

    @property
    def u_hat_(self):
        """array: spatial map of the dictionary. If `rank1 = False`, this is an
        approximation of the dictionary obtained through svd.

        , shape (n_atoms, n_channels)
        """
        return self.uv_hat_[:, :self.n_channels_]

    @property
    def v_hat_(self):
        """array: temporal patterns of the dictionary. If `rank1 = False`, this
        is an approximation of the dictionary obtained through svd.

        shape (n_atoms, n_times_atom)
        """
        return self.uv_hat_[:, self.n_channels_:]

    @property
    def z_hat_(self):
        """array: Sparse code associated to the signals used to fit the model.

        shape (n_trials, n_atoms, n_times_valid)
        """
        self._check_fitted()
        return self._z_hat

    @property
    def pobj_(self):
        """list: Objective function value at each step of the alternate minimization.
        """
        self._check_fitted()
        return self._pobj

    @property
    def times_(self):
        """list: Cumulative time for each iteration of the coordinate descent.
        """
        self._check_fitted()
        return self._times


class BatchCDL(ConvolutionalDictionaryLearning):
    _default = {}
    _default.update(DEFAULT)
    _default['desc'] = "Batch algorithm for convolutional dictionary learning"
    _default['algorithm'] = "    Batch algorithm\n"
    __doc__ = DOC_FMT.format(**_default)

    def __init__(self, n_atoms, n_times_atom, reg=0.1, n_iter=60, n_jobs=1,
                 solver_z='lgcd', solver_z_kwargs={}, unbiased_z_hat=False,
                 solver_d='alternate_adaptive', solver_d_kwargs={},
                 rank1=True, window=False, uv_constraint='separate',
                 lmbd_max='scaled', eps=1e-10, D_init=None, D_init_params={},
                 verbose=10, random_state=None, sort_atoms=False):
        super().__init__(
            n_atoms, n_times_atom, reg=reg, n_iter=n_iter,
            solver_z=solver_z, solver_z_kwargs=solver_z_kwargs,
            rank1=rank1, window=window, uv_constraint=uv_constraint,
            unbiased_z_hat=unbiased_z_hat, sort_atoms=sort_atoms,
            solver_d=solver_d, solver_d_kwargs=solver_d_kwargs,
            eps=eps, D_init=D_init, D_init_params=D_init_params,
            algorithm='batch', lmbd_max=lmbd_max, raise_on_increase=True,
            loss='l2', use_sparse_z=False, n_jobs=n_jobs, verbose=verbose,
            callback=None, random_state=random_state, name="BatchCDL")


class GreedyCDL(ConvolutionalDictionaryLearning):
    _default = {}
    _default.update(DEFAULT)
    _default['desc'] = ("Greedy batch algorithm for convolutional dictionary "
                        "learning")
    _default['algorithm'] = "    Greedy batch algorithm\n"
    __doc__ = DOC_FMT.format(**_default)

    def __init__(self, n_atoms, n_times_atom, reg=0.1, n_iter=60, n_jobs=1,
                 solver_z='lgcd', solver_z_kwargs={}, unbiased_z_hat=False,
                 solver_d='alternate_adaptive', solver_d_kwargs={},
                 rank1=True, window=False, uv_constraint='separate',
                 lmbd_max='scaled', eps=1e-10, D_init=None, D_init_params={},
                 verbose=10, random_state=None, sort_atoms=False):
        super().__init__(
            n_atoms, n_times_atom, reg=reg, n_iter=n_iter,
            solver_z=solver_z, solver_z_kwargs=solver_z_kwargs,
            rank1=rank1, window=window, uv_constraint=uv_constraint,
            unbiased_z_hat=unbiased_z_hat, sort_atoms=sort_atoms,
            solver_d=solver_d, solver_d_kwargs=solver_d_kwargs,
            eps=eps, D_init=D_init, D_init_params=D_init_params,
            algorithm='greedy', lmbd_max=lmbd_max, raise_on_increase=True,
            loss='l2', use_sparse_z=False, n_jobs=n_jobs, verbose=verbose,
            callback=None, random_state=random_state, name="GreedyCDL")



import os
import ctypes
import traceback


def get_calling_script():
    """Get the calling script from the traceback stack"""
    stack = traceback.extract_stack()

    script_path = None
    for trace in stack:
        if trace[2] == '<module>':
            script_path = trace[0]
    if script_path is None:
        for trace in stack:
            if '/run_' in trace[0]:
                script_path = trace[0]
    if script_path is None:
        script_path = stack[-1][0]  # default

    script_name = os.path.basename(script_path)
    if script_name[:14] == '<ipython-input':
        script_name = '<ipython>'
    if script_name[-3:] == '.py':
        script_name = script_name[:-3]

    return script_name


def positive_hash(obj):
    return ctypes.c_size_t(hash(obj)).value



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

import numpy as np
cimport numpy as cnp
cnp.import_array()


cdef compute_ztX(object[:] zi_data,
                 object[:] zi_rows,
                 cnp.ndarray[double, ndim=2] Xi,
                 cnp.ndarray[double, ndim=3] ztX):

    cdef int k, t
    cdef double z
    cdef int n_times_atom = ztX.shape[2]

    for k, (zk, tk) in enumerate(zip(zi_data, zi_rows)):
        for z, t in zip(zk, tk):
            ztX[k, :, :] += z * Xi[:, t:t + n_times_atom]

    return ztX


def _fast_compute_ztX(z, X):
    """
    z.shape = n_atoms, n_trials, n_times - n_times_atom + 1)
    X.shape = n_trials, n_channels, n_times
    ztX.shape = n_atoms, n_channels, n_times_atom
    """
    n_trials, n_channels, n_times = X.shape
    assert n_trials == len(z)
    n_atoms, n_times_valid = z[0].shape
    n_times_atom = n_times - n_times_valid + 1
    ztX = np.zeros(shape=(n_atoms, n_channels, n_times_atom))
    for zi, Xi in zip(z, X):
        ztX = compute_ztX(zi.data, zi.rows, Xi, ztX)

    return ztX
